# -*- coding: utf-8 -*-
#python 3 compat
#http://python-future.org/compatible_idioms.html
from __future__ import print_function

import sys
import os
import re
import pprint

from .common import find_umi_groups, parse_extended_read_name

import logging
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
sh = logging.StreamHandler()
sh.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
log.addHandler(sh)
import time

#use pandas to read the CSV file and write output files
import pandas as pd
#for defaultdict + sorting
import collections
import operator

#use python's argumentparser to support command line parameters like --probe=62
import argparse

#for pileup
import pysam
#for getting ref base 
import pyfaidx
#for debugging umi distance
import distance

#import reader script for reading probe design
from .reader import read_new_probe_design

class ConsensusFilterException(Exception):
    def __init__(self, filter_column):
        self.filter_column = filter_column

def process_file(design, input, output,
        min_mapq, min_consensus_count,
        include_primers, use_naive_groups, ignore_groups,
        ignore_group_mismatch = False,
        debug = False):
    assert design is not None, 'design file missing'
    assert input is not None, 'input file missing'
    assert output is not None, 'output file missing'

    design = read_new_probe_design(design)

    if include_primers:
        probe_column_start = 'probe_start'
        probe_column_end = 'probe_end'
    else:
        probe_column_start = 'target_start'
        probe_column_end = 'target_end'
    log.info('Using columns %s / %s to determine target regions', probe_column_start, probe_column_end)

    log.info('Processing file %s', input)
    if not os.path.isfile(input):
        sys.exit('input does not exist: %s' % input)

    def process_aln_pile(outbam, alns_by_umi):
        #determine counts per umi
        umi_counts = dict( [(umi, len(umi_alns)) for umi, umi_alns in alns_by_umi.items() ]  )
        #group these raw umis into umi groups
        umi_to_group = find_umi_groups( umi_counts )

        group_firstread = {}
        group_qnames = collections.defaultdict(list)
        group_cigars = collections.defaultdict(collections.Counter())
        group_max_mapqual = collections.defaultdict(int)
        group_bases = {}
        group_max_basequal = {}

        for umi, umi_alns in alns_by_umi.items():
            umi_group = umi_to_group[umi]

            for aln in umi_alns:
                #we also need to handle different mate starts
                umi_group_mate = '%d_%d_%d' % (umi_group, aln.next_reference_id, aln.next_reference_start)
                if not umi_group_mate in group_firstread:
                    group_firstread[umi_group_mate] = aln
                    group_bases[umi_group_mate] = collections.defaultdict(dict)
                    group_max_basequal[umi_group_mate] = collections.defaultdict(dict)

                group_qnames[umi_group_mate].append(aln.query_name)
                group_cigars[umi_group_mate][aln.cigarstring] += 1
                group_max_mapqual[umi_group_mate] = max(group_max_mapqual[umi_group_mate], aln.mapping_quality)

                #go through read base by base to form consensus
                for index in range( len(aln.query_sequence) ):
                    group_bases[umi_group_mate][index][aln.query_sequence[index]] += 1
                    group_max_basequal[umi_group_mate][index][aln.query_sequence[index]] = max(
                        group_max_basequal[umi_group_mate][index][aln.query_sequence[index]],
                        aln.query_qualities[index]
                        )

        for umi_group_mate, aln in group_firstread.items():
            aln.query_name = 'group__%s' % umi_group_mate
            aln.set_tag('rc', len(group_qnames[umi_group_mate]))
            aln.set_tag('rn', ';'.join(group_qnames[umi_group_mate]))

            aln.cigarstring = group_cigars[umi_group_mate].most_common(1)[0][0]
            aln.mapping_quality = group_max_mapqual[umi_group_mate]

            #TODO: do we need to do something about the flags?

            #determine most common base at each position
            my_bases = [counts.most_common(1)[0][0] for counts in group_bases[umi_group_mate]]
            my_sequence = ''.join( my_bases )
            #determine highest quality score for most common base at each position
            my_qualities = [group_maxqual[umi_group_mate][base] for base in my_bases]

            #set sequence and qualities
            #note: need to set seq before qual
            aln.query_sequence = my_sequence
            aln.query_qualities = my_qualities

    with tempfile.TemporaryFile() as tmp:
        with pysam.AlignmentFile(input, 'rb') as inbam:
            with pysam.AlignmentFile(tmp, 'wb', template = inbam) as outbam:
                t_shown = t_start = time.time()

                counters = collections.Counter()
                mates_of_invalid_reads = set()
                current_location = (None, None)
                current_alns = collections.defaultdict(list)

                #doc: If until_eof is True, all reads from the current file position will be returned in order as they are within the file. Using this option will also fetch unmapped reads.
                for aln in inbam.fetch(until_eof = True):
                    rindex = 1 if aln.is_read1 else 2 if aln.is_read2 else None
                    assert rindex is not None
                    other_rindex = 3 - rindex

                    #only count rows after this to fulfil assumption later
                    counters['alignments_in'] += 1

                    try:
                        #supplementary alignments mess up the assumption that each mate has one other mate
                        #so we ignore them for now
                        if aln.is_supplementary:
                            raise ConsensusFilterException('is_supplementary')

                        #we also don't want unmapped reads here
                        if aln.is_unmapped:
                            raise ConsensusFilterException('is_unmapped')

                        #and no unmapped mates
                        if aln.mate_is_unmapped:
                            raise ConsensusFilterException('mate_is_unmapped')

                        if debug:
                            print(aln)

                        #parse info from read name (should be bowtie2 compatible)
                        read_name, read_probe, read_umi = parse_extended_read_name(aln.query_name)

                        #make sure we have a valid read pair
                        is_valid_pair = True
                        is_valid_pair = is_valid_pair and aln.is_reverse != aln.mate_is_reverse #should be opposite
                        is_valid_pair = is_valid_pair and aln.reference_id == aln.next_reference_id #should be on same chr
                        if debug:
                            print('is_valid_pair =', is_valid_pair)

                        #and no unmapped mates
                        if not is_valid_pair:
                            raise ConsensusFilterException('not_valid_pair')

                        #check probe data
                        probe_data = design.loc[pname, :]
                        matches_probe = aln.reference_name == probe_data['chr']
                        if matches_probe:
                            #TODO: this may be the wrong way round!
                            my_column = probe_column_start
                            if aln.is_reverse or (not aln.is_reverse and aln.is_read2):
                                matches_probe = matches_probe and aln.reference_end + 1 == probe_data[probe_column_end]
                            else:
                                matches_probe = matches_probe and aln.reference_start + 1 == probe_data[probe_column_start]
                        if debug:
                            print('matches_probe =', matches_probe)

                        #and no unmapped mates
                        if not matches_probe:
                            raise ConsensusFilterException('mismatches_probe')

                        #update flags if this is a mate of a read that has been filtered
                        my_tuple = (aln.query_name, aln.reference_id, aln.reference_start)
                        if my_tuple in mates_of_invalid_reads:
                            mates_of_invalid_reads.remove( my_tuple )
                            raise ConsensusFilterException('mate_filtered')

                        #is this a new pile of alignments?
                        if current_location[0] != aln.reference_id or current_location[1] != aln.reference_start:
                            process_aln_pile(outbam, current_alns)

                            current_alns = collections.defaultdict(list)
                            current_location = (aln.reference_id, aln.reference_start)

                        assert read_umi is not None and len(read_umi) > 0
                        read_umi_bytes = read_umi.encode('utf-8')
                        current_alns[read_umi_bytes].append(aln)
                    except ConsensusFilterException as ex:
                        #if the row is filtered, we raise an exception with message being the corresponding counter column
                        counters[ex.filter_column] += 1

                        #keep track of mates we still need to filter out
                        if ex.filter_column != 'mate_filtered':
                            mates_of_invalid_reads.add( (aln.query_name, aln.next_reference_id, aln.next_reference_start) )

                    #print some stats
                    t_now = time.time()
                    if n_rows > 0 and (n_rows % 10000 == 0 or (t_now - t_shown > 60)):
                        t_shown = t_now
                        log.info("processed %d alignment rows - %.f sec elapsed, %.4f sec/pair, %.1f row/sec",
                            counters['alignments_in'],
                            t_now - t_start,
                            (t_now - t_start) / counters['alignments_in'],
                            counters['alignments_in'] / (t_now - t_start))

                #run this one final time to catch the last pile
                process_aln_pile(outbam, current_alns)

            #summary stats
            log.info('processed %d alignment rows (reads) total.', counters['alignments_in'])

        log.info('Re-processing file to remove invalid pairs...', counters['alignments'])
        with pysam.AlignmentFile(tmp, 'rb') as inbam:
            with pysam.AlignmentFile(output, 'wb', template = inbam) as outbam:
                for aln in inbam.fetch(until_eof = True):
                    counters['alignments_first_pass'] += 1

                    my_tuple = (aln.query_name, aln.reference_id, aln.reference_start)
                    if my_tuple in mates_of_invalid_reads:
                        mates_of_invalid_reads.remove( my_tuple )
                        counters['mate_filtered_second_pass'] += 1
                    else:
                        counters['alignments_second_pass'] += 1
                        outbam.write(aln)

        log.info('read %d rows from tmp file, wrote %d rows.', counters['alignments_first_pass'], counters['alignments_second_pass'])
        print(output)

def main():
    log.info('Called with arguments: "%s"', '" "'.join(sys.argv))
    
    #parse the arguments, which will be available as properties of args (e.g. args.probe)
    parser = argparse.ArgumentParser()
    #specify parameters
    parser.add_argument("--aggregate", help="folder with processed CSV files to aggregate")
    parser.add_argument("-d", "--design", help="CSV file with probes, arms sequences and locations")
    parser.add_argument("-i", "--input", help="input bam file name")
    parser.add_argument("-o", "--output", help="output bam file name")
    parser.add_argument("--use-naive-groups", help="use UMIs from um tag instead of UG tags", action='store_true')
    parser.add_argument("--ignore-groups", help="ignore the UMI group entirely", action='store_true')
    parser.add_argument("--ignore-group-mismatch", help="ignore group mismatches (to deal with old alignments that have separate UMIs)", action='store_true')
    parser.add_argument("--min-consensus-count", help="minimum number of consistent read pairs supporting a UMI", default=2, type=int)
    parser.add_argument("--min-mapq", help="minimum mapping quality (remove read alignments with either read having mapq < X)", default=20, type=int)
    parser.add_argument("--include-primers", help="include primers when determining on/off target", action='store_true')

    parser.add_argument("--debug", help="enable debug output", action='store_true')
    args = parser.parse_args()

    process_file(args.design, args.input, args.output,
        args.min_mapq, args.min_consensus_count,
        args.include_primers, args.use_naive_groups, args.ignore_groups,
        args.ignore_group_mismatch,
        debug = args.debug)

    return 0

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    sys.exit(main())


# make sure we can import from package directory
import sys, os
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import unittest
import string
import math
import io
import amplimap.common
import amplimap.reader
from amplimap.reader import *

class Reader(unittest.TestCase):
    def test_mipgen_convert(self):
        # single change (case from MIPGEN practice files)
        f = io.StringIO(""">mip_key	logistic_score	chr	ext_probe_start	ext_probe_stop	ext_probe_copy	ext_probe_sequence	lig_probe_start	lig_probe_stop	lig_probe_copy	lig_probe_sequence	mip_scan_start_position	mip_scan_stop_position	scan_target_sequence	mip_sequence	feature_start_position	feature_stop_position	probe_strand	failure_flags	mip_name
5:1253823-1253984/17,23/+	0.798319	5	1253823	1253839	2	TGGCTGTGGGCGGGTGG	1253962	1253984	1	CAGACTCCGTTCCAGAAGAGGCC	1253840	1253961	CCATCAGTCCAGGATGGTCTTGAAGTCTGAGGGCAGTGCCGGGTTGGCTGCGGCCTCCAGGGCAGTCAGCGTCGTCCCCGGGAGCTTCCGACTCAGCTGCGTCTGGGCTGCGGGGCCAAAAT	CAGACTCCGTTCCAGAAGAGGCCCTTCAGCTTCCCGATATCCGACGGTAGTGTNNNNNTGGCTGTGGGCGGGTGG	1253842	1253946	+	000	TERT/NM_001193376_0001
5:1282608-1282769/26,19/-	0.909748	5	1282744	1282769	1	TCTGGATGTGGCATGTCCTTCTCGTT	1282608	1282626	1	GTCACGGAGACCACGTTTC	1282627	1282743	TAAGGGGTTGGCTGTGTTCCGGCCGCAGAGCACCGTCTGCGTGAGGAGATCCTGGCCAAGTTCCTGCACTGGCTGATGAGTGTGTACGTCGTCGAGCTGCTCAGGTCTTTCTTTTAT	GTCACGGAGACCACGTTTCCTTCAGCTTCCCGATATCCGACGGTAGTGTNNNNNTCTGGATGTGGCATGTCCTTCTCGTT	1282543	1282739	-	000	TERT/NM_001193376_0028_SNP_a
5:1282608-1282769/26,19/-	0.909748	5	1282744	1282769	1	TCTGGATGTGGCATCTCCTTCTCGTT	1282608	1282626	1	GTCACGGAGACCACGTTTC	1282627	1282743	TAAGGGGTTGGCTGTGTTCCGGCCGCAGAGCACCGTCTGCGTGAGGAGATCCTGGCCAAGTTCCTGCACTGGCTGATGAGTGTGTACGTCGTCGAGCTGCTCAGGTCTTTCTTTTAT	GTCACGGAGACCACGTTTCCTTCAGCTTCCCGATATCCGACGGTAGTGTNNNNNTCTGGATGTGGCATCTCCTTCTCGTT	1282543	1282739	-	000	TERT/NM_001193376_0028_SNP_b""")
        probes = amplimap.reader.read_and_convert_mipgen_probes(f, sep='\t')

        assert len(probes) == 2
        assert probes.iloc[1]['id'] == 'TERT/NM_001193376_0028'
        assert probes.iloc[1]['second_primer_5to3'] == 'TCTGGATGTGGCAT.TCCTTCTCGTT'
        
        # multiple changes
        f = io.StringIO(""">mip_key	logistic_score	chr	ext_probe_start	ext_probe_stop	ext_probe_copy	ext_probe_sequence	lig_probe_start	lig_probe_stop	lig_probe_copy	lig_probe_sequence	mip_scan_start_position	mip_scan_stop_position	scan_target_sequence	mip_sequence	feature_start_position	feature_stop_position	probe_strand	failure_flags	mip_name
5:1253823-1253984/17,23/+	0.798319	5	1253823	1253839	2	TGGCTGTGGGCGGGTGG	1253962	1253984	1	CAGACTCCGTTCCAGAAGAGGCC	1253840	1253961	CCATCAGTCCAGGATGGTCTTGAAGTCTGAGGGCAGTGCCGGGTTGGCTGCGGCCTCCAGGGCAGTCAGCGTCGTCCCCGGGAGCTTCCGACTCAGCTGCGTCTGGGCTGCGGGGCCAAAAT	CAGACTCCGTTCCAGAAGAGGCCCTTCAGCTTCCCGATATCCGACGGTAGTGTNNNNNTGGCTGTGGGCGGGTGG	1253842	1253946	+	000	TERT/NM_001193376_0001
5:1282608-1282769/26,19/-	0.909748	5	1282744	1282769	1	AAAGGATGTGGCATGTCCTTCTCGTT	1282608	1282626	1	GTCACGGAGACCACGTTTC	1282627	1282743	TAAGGGGTTGGCTGTGTTCCGGCCGCAGAGCACCGTCTGCGTGAGGAGATCCTGGCCAAGTTCCTGCACTGGCTGATGAGTGTGTACGTCGTCGAGCTGCTCAGGTCTTTCTTTTAT	GTCACGGAGACCACGTTTCCTTCAGCTTCCCGATATCCGACGGTAGTGTNNNNNTCTGGATGTGGCATGTCCTTCTCGTT	1282543	1282739	-	000	TERT/NM_001193376_0028_SNP_a
5:1282608-1282769/26,19/-	0.909748	5	1282744	1282769	1	CCCGGATGTGGCATCTCCTTCTCGTT	1282608	1282626	1	GTCACGGAGACCACGTTTC	1282627	1282743	TAAGGGGTTGGCTGTGTTCCGGCCGCAGAGCACCGTCTGCGTGAGGAGATCCTGGCCAAGTTCCTGCACTGGCTGATGAGTGTGTACGTCGTCGAGCTGCTCAGGTCTTTCTTTTAT	GTCACGGAGACCACGTTTCCTTCAGCTTCCCGATATCCGACGGTAGTGTNNNNNTCTGGATGTGGCATCTCCTTCTCGTT	1282543	1282739	-	000	TERT/NM_001193376_0028_SNP_b""")
        probes = amplimap.reader.read_and_convert_mipgen_probes(f, sep='\t')

        assert len(probes) == 2
        assert probes.iloc[1]['id'] == 'TERT/NM_001193376_0028'
        assert probes.iloc[1]['second_primer_5to3'] == '...GGATGTGGCAT.TCCTTCTCGTT'
        
        # different IDs, no changes
        f = io.StringIO(""">mip_key	logistic_score	chr	ext_probe_start	ext_probe_stop	ext_probe_copy	ext_probe_sequence	lig_probe_start	lig_probe_stop	lig_probe_copy	lig_probe_sequence	mip_scan_start_position	mip_scan_stop_position	scan_target_sequence	mip_sequence	feature_start_position	feature_stop_position	probe_strand	failure_flags	mip_name
5:1253823-1253984/17,23/+	0.798319	5	1253823	1253839	2	TGGCTGTGGGCGGGTGG	1253962	1253984	1	CAGACTCCGTTCCAGAAGAGGCC	1253840	1253961	CCATCAGTCCAGGATGGTCTTGAAGTCTGAGGGCAGTGCCGGGTTGGCTGCGGCCTCCAGGGCAGTCAGCGTCGTCCCCGGGAGCTTCCGACTCAGCTGCGTCTGGGCTGCGGGGCCAAAAT	CAGACTCCGTTCCAGAAGAGGCCCTTCAGCTTCCCGATATCCGACGGTAGTGTNNNNNTGGCTGTGGGCGGGTGG	1253842	1253946	+	000	TERT/NM_001193376_0001
5:1282608-1282769/26,19/-	0.909748	5	1282744	1282769	1	AAAGGATGTGGCATGTCCTTCTCGTT	1282608	1282626	1	GTCACGGAGACCACGTTTC	1282627	1282743	TAAGGGGTTGGCTGTGTTCCGGCCGCAGAGCACCGTCTGCGTGAGGAGATCCTGGCCAAGTTCCTGCACTGGCTGATGAGTGTGTACGTCGTCGAGCTGCTCAGGTCTTTCTTTTAT	GTCACGGAGACCACGTTTCCTTCAGCTTCCCGATATCCGACGGTAGTGTNNNNNTCTGGATGTGGCATGTCCTTCTCGTT	1282543	1282739	-	000	TERT/NM_001193376_0028_SNP_a
5:1282608-1282769/26,19/-	0.909748	5	1282744	1282769	1	CCCGGATGTGGCATCTCCTTCTCGTT	1282608	1282626	1	GTCACGGAGACCACGTTTC	1282627	1282743	TAAGGGGTTGGCTGTGTTCCGGCCGCAGAGCACCGTCTGCGTGAGGAGATCCTGGCCAAGTTCCTGCACTGGCTGATGAGTGTGTACGTCGTCGAGCTGCTCAGGTCTTTCTTTTAT	GTCACGGAGACCACGTTTCCTTCAGCTTCCCGATATCCGACGGTAGTGTNNNNNTCTGGATGTGGCATCTCCTTCTCGTT	1282543	1282739	-	000	TERT/NM_001193376_0028XXX_SNP_b""")
        probes = amplimap.reader.read_and_convert_mipgen_probes(f, sep='\t')

        assert len(probes) == 3
        assert probes.iloc[1]['id'] == 'TERT/NM_001193376_0028'
        assert probes.iloc[1]['second_primer_5to3'] == 'AAAGGATGTGGCATGTCCTTCTCGTT'
        assert probes.iloc[2]['id'] == 'TERT/NM_001193376_0028XXX'
        assert probes.iloc[2]['second_primer_5to3'] == 'CCCGGATGTGGCATCTCCTTCTCGTT'

if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module provides the ``amplimap.run.main()`` method called from the `amplimap` command-line executable,
as well as some helper functions for reading and checking the config files.
"""

import os
import sys

import snakemake
import argparse
import yaml

from .version import __title__, __version__
from .reader import AmplimapReaderException, read_new_probe_design, process_probe_design, read_and_convert_mipgen_probes, read_and_convert_heatseq_probes, read_targets, read_snps_txt

def check_config_keys(default_config, my_config, path = []):
    """Recursively check that config keys provided in my_config also exist in default_config (ignoring 'paths' and 'clusters')."""
    differences = []
    for key, value in my_config.items():
        if key in default_config:
            if isinstance(default_config[key], dict) == isinstance(value, dict):
                if isinstance(default_config[key], dict):
                    #both are dicts, keep checking
                    #but don't check paths/cluster because they can contain custom values
                    if not (len(path) == 0 and key in ['paths', 'clusters']):
                        differences += check_config_keys(default_config[key], value, path + [key])
                else:
                    #we're done here
                    pass
            else:
                differences.append(path + [key])
                #raise Exception('Config setting {} is invalid\n'.format(':'.join(path+[key])))
        else:
            differences.append(path + [key])
    return differences

def compare_config_dicts(my_config, used_config, path = []):
    """Recursively search for differences in values between two dicts."""
    differences = []
    for key, value in my_config.items():
        if key in used_config:
            if isinstance(value, dict):
                differences += compare_config_dicts(value, used_config[key], path + [key])
            else:
                if value != used_config[key]:
                    differences.append(path + [key])
        else:
            #key does not exist in used config
            sys.stderr.write('Warning - config key {} not set in used config\n'.format(':'.join(path+[key])))
    return differences

def read_config_file(print_config, path):
    """
    Helper function to read a config file, if it exists.
    Always returns a dict, but it may be empty.
    """
    this_config = {}
    if os.path.isfile(path):
        if print_config:
            sys.stderr.write('Reading additional configuration file: {}\n'.format(path))
        with open(path, 'r') as config_file:
            this_config = yaml.safe_load(config_file.read())

        #make sure we always get an empty dict, yaml.safe_load may give us None for an empty file
        if this_config is None:
            this_config = {}

    return this_config

def main(argv = None):
    """
    Main function for the ``amplimap`` executable. This function:

    - parses command line arguments
    - reads, merges and checks each of these config files, if they exist:
        + ``config_default.yaml`` in the amplimap package
        + ``/etc/amplimap/VERSION/config.yaml`` (where VERSION is the amplimap version)
        + ``$AMPLIMAP_CONFIG``
        + ``config.yaml`` in the working directory
    - checks for an existing analysis directory (and compares the amplimap version used to create it)
    - adds its own parent directory to the config file (to be inserted back into the python path inside Snakemake)
    - creates an analysis directory
    - writes ``config_used.yaml`` to the new analysis directory
    - creates a ``cluster_log`` directory (if running in cluster mode)
    - launches Snakemake, using the amplimap Snakefile, ``config_used.yaml`` as the config file and cluster parameters as specified in the command line arguments and config.
    """
    try:
        basedir = os.path.dirname(os.path.realpath(__file__))

        # parse the arguments, which will be available as properties of args (e.g. args.probe)
        parser = argparse.ArgumentParser(
            description = "amplimap v{} - amplicon mapping and analysis pipeline".format(__version__),
            formatter_class = argparse.ArgumentDefaultsHelpFormatter)
        # specify parameters
        parser.add_argument("-v", "--version", help="print version and exit", action="store_true")
        parser.add_argument("--basedir", help="print basedir and exit", action="store_true")
        parser.add_argument("--print-config", help="print configuration (including global and local settings) and exit", action="store_true")
        parser.add_argument("-r", "--run", help="actually run (will perform a dry run otherwise)", action="store_true")
        parser.add_argument("--resume", help="resume analysis in existing analysis directory", action="store_true")
        parser.add_argument("--cluster", help="specify a cluster type defined in your configuration files to run jobs on cluster.")
        parser.add_argument("--skip-file-check", help="skip check for changes in input files when resuming (not recommended)", action="store_true")
        parser.add_argument("--unlock", help="unlock working directory (Snakemake parameter)", action="store_true")
        parser.add_argument("--working-directory", help="path to the working directory", default=".")
        parser.add_argument("--ncores", help="number of local cores to run in parallel (only applies if --cluster is NOT set)", default=1, type=int)
        parser.add_argument("--njobs", help="number of cluster jobs to run in parallel (only applies if --cluster is set)", default=10, type=int)
        parser.add_argument("--latency-wait", help="How long to wait for output files to appear after job completes. Increase this if you get errors about missing output files. (Snakemake parameter)", default=5, type=int)
        parser.add_argument("--snakemake-args", help="For debugging: Extra arguments to the snakemake function (comma-separated key=value pairs - eg. 'printreason=True')")
        parser.add_argument("--debug", help="debug mode", action="store_true")
        # parser.add_argument("--debug-dag", help="debug DAG", action="store_true")
        parser.add_argument("TARGET", help="targets to run (eg. pileups variants coverages)", nargs="*")
        if argv is None:
            args = parser.parse_args()
        else:
            args = parser.parse_args(argv)

        if args.debug:
            print('Incoming argv: {}'.format(str(argv)))
            print('Targets: {}'.format(str(args.TARGET)))

        if args.version:
            print('{} {}'.format(__title__, __version__))
            return 0

        if args.basedir:
            print(basedir)
            return 0

        # read base config to know which parameters etc are allowed
        default_config = read_config_file(args.print_config, os.path.join(basedir, 'config_default.yaml'))
        if not default_config:
            raise Exception('config_default.yaml file missing!')

        # add undocumented config keys to make sure these don't raise an error
        for key in ['include_gbrowse_links', 'include_exon_distance', 'include_score']:
            if not key in default_config['annotate']:
                default_config['annotate'][key] = False

        # override with data from /etc/amplimap, if exists
        etc_config = read_config_file(args.print_config, '/etc/amplimap/%s/config.yaml' % __version__)

        # override with data from $AMPLIMAP_CONFIG, if exists
        env_config = {}
        try:
            env_config = read_config_file(args.print_config, os.environ['AMPLIMAP_CONFIG'])
        except KeyError:
            pass

        # read local config
        local_config = read_config_file(args.print_config, os.path.join(args.working_directory, 'config.yaml'))
        if not local_config:
            if args.print_config:
                sys.stderr.write('No local config.yaml found, using default configuration.\n')

        # merge configs together
        config = default_config
        for my_config in [etc_config, env_config, local_config]:
            # check that all settings actually exist
            differences = check_config_keys(default_config, my_config)

            # allow custom tools
            allowed_tools = set(default_config['tools'].keys())
            if 'tools' in my_config:
                allowed_tools.update(my_config['tools'].keys())
            differences = [
                d for d in differences
                if not (len(d) == 2 and d[0] in ['modules'] and d[1] in allowed_tools)
                and d[0] != 'tools'
            ]

            if len(differences) > 0:
                sys.stderr.write('Your configuration file(s) contain unknown or invalid settings:\n')
                for diff in differences:
                    sys.stderr.write('\t- {}\n'.format(':'.join(diff)))
                sys.stderr.write('Please check their spelling and location and try again.\n')
                return 1

            snakemake.utils.update_config(config, my_config)

        # check basic config
        aligners = ['naive', 'bwa', 'bowtie2', 'star']  # allowed values for the aligner
        # add custom tools
        for tool_name, tool_config in config['tools'].items():
            if 'align_command' in tool_config:
                aligners.append(tool_name)
        if not config['align']['aligner'] in aligners:
            raise Exception('align: aligner must be one of {}!'.format(','.join(aligners)))

        callers = ['gatk', 'platypus', 'wecall']  # allowed values for the variant caller
        # add custom tools
        for tool_name, tool_config in config['tools'].items():
            if 'call_command' in tool_config:
                callers.append(tool_name)
        if not config['variants']['caller'] in callers:
            raise Exception('variants: caller must be one of {}!'.format(','.join(callers)))

        if config['parse_reads']['quality_trim_threshold'] != False:
            if not isinstance(config['parse_reads']['quality_trim_threshold'], float):
                raise Exception('quality_trim_threshold must be a decimal number!')
            if not config['parse_reads']['quality_trim_threshold'] > 0 and config['parse_reads']['quality_trim_threshold'] < 1:
                raise Exception('quality_trim_threshold must be either "false" or above 0 and below 1!')

        if not (config['general']['umi_min_consensus_percentage'] >= 0 and config['general']['umi_min_consensus_percentage'] <= 100):
            raise Exception('umi_min_consensus_percentage must be between 0 and 100')

        if not (config['parse_reads']['min_percentage_good'] >= 0 and config['parse_reads']['min_percentage_good'] <= 100):
            raise Exception('min_percentage_good must be between 0 and 100')

        if not (config['parse_reads']['umi_one'] >= 0 and config['parse_reads']['umi_two'] >= 0):
            raise Exception('umi_one and umi_two must be 0 or greater')

        if config['annotate']['annovar']['protocols'].count(',') != config['annotate']['annovar']['operations'].count(','):
            raise Exception('The number of comma-separated protocols and operations under `annotate: annovar:` must match!')

        #if we don't have UMIs (either on reads or as bam tag) we definitely have to ignore them
        #this makes it possible to "auto-detect" whether we need to ignore_umis or not
        if (config['parse_reads']['umi_one'] + config['parse_reads']['umi_two'] == 0) and config['general']['umi_tag_name'] == "":
            config['general']['ignore_umis'] = True

        #check we have proper paths
        if not config['general']['genome_name'] in config['paths'] or not isinstance(config['paths'][config['general']['genome_name']], dict):
            raise Exception('Could not find list of paths for genome_name: "{}". Please add the paths to your default configuration or your local config.yaml file.'.format(config['general']['genome_name']))

        for name, path in config['paths'][config['general']['genome_name']].items():
            if path.startswith('/PATH/TO/'):
                raise Exception('Path for {} reference is set to {}, which is probably incorrect. Please set the correct path in your default configuration or your local config.yaml file, or leave it empty.'.format(
                    name, path))

        if args.print_config:
            yaml.dump(config, sys.stdout, default_flow_style=False)
            return 0

        #do some basic checks
        assert os.path.isdir(args.working_directory), 'working directory does not exist'

        #check for one (and only one) input directory
        input_directory = None
        input_directory_count = 0
        input_directories = ['reads_in', 'unmapped_bams_in', 'mapped_bams_in', 'bams_in']
        for input_name in input_directories:
            if os.path.isdir(os.path.join(args.working_directory, input_name)):
                input_directory_count += 1
                input_directory = input_name
        if input_directory_count < 1:
            raise Exception(
                'An input directory (one of: %s) needs to exist. Please see the documentation for the appropriate directory to use and place your sequencing data there.'
                % (', '.join(input_directories))
            )
        elif input_directory_count > 1:
            raise Exception(
                'More than one of the possible input directories (%s) exists. Please only provide a single input directory with all your data.'
                % (', '.join(input_directories))
            )

        if input_directory in ['unmapped_bams_in', 'mapped_bams_in']:
            if not config['general']['use_raw_reads']:
                raise Exception(
                    'general: use_raw_reads needs to be set to true when using %s for input.'
                    % (input_directory)
                )

        #check input files
        sys.stderr.write('Checking input files...\n')
        if os.path.isfile(os.path.join(args.working_directory, 'probes.csv')):
            read_new_probe_design(os.path.join(args.working_directory, 'probes.csv'), reference_type = 'genome')
        if os.path.isfile(os.path.join(args.working_directory, 'probes_mipgen.csv')):
            process_probe_design(read_and_convert_mipgen_probes(os.path.join(args.working_directory, 'probes_mipgen.csv')))
        if os.path.isfile(os.path.join(args.working_directory, 'picked_mips.txt')):
            process_probe_design(read_and_convert_mipgen_probes(os.path.join(args.working_directory, 'picked_mips.txt'), sep='\t'))
        if os.path.isfile(os.path.join(args.working_directory, 'probes_heatseq.tsv')):
            process_probe_design(read_and_convert_heatseq_probes(os.path.join(args.working_directory, 'probes_heatseq.tsv')))
        if os.path.isfile(os.path.join(args.working_directory, 'targets.csv')):
            #note: this will fail on overlapping targets
            read_targets(os.path.join(args.working_directory, 'targets.csv'), check_overlaps=True, reference_type = 'genome', file_type = 'csv')
        if os.path.isfile(os.path.join(args.working_directory, 'targets.bed')):
            #note: this will fail on overlapping targets
            read_targets(os.path.join(args.working_directory, 'targets.bed'), check_overlaps=True, reference_type = 'genome', file_type = 'bed')
        if os.path.isfile(os.path.join(args.working_directory, 'snps.txt')):
            read_snps_txt(os.path.join(args.working_directory, 'snps.txt'), reference_type = 'genome')

        #this will be used to (very hackily) make sure amplimap can be imported as amplimap.xxx
        #by adding the parent dir to the top of sys.path in the Snakefile
        config['general']['amplimap_parent_dir'] = os.path.dirname(basedir)

        #check if analysis dir exists already
        analysis_dir = os.path.join(args.working_directory, 'analysis')
        configfile = os.path.join(analysis_dir, 'config_used.yaml')
        used_versions_path = os.path.join(analysis_dir, 'versions.yaml')

        #the analysis dir may exist just because we did a dry run, but once the versions exist we actually executed snakemake!
        if os.path.exists(analysis_dir) and os.path.exists(used_versions_path):
            if not args.resume:
                raise Exception('An analysis directory already exists. Please rename it or set --resume to reuse it and possibly overwrite existing files.')
            else:
                #check version
                if os.path.isfile(used_versions_path):
                    with open(used_versions_path, 'r') as used_versions_file:
                        used_versions = yaml.safe_load(used_versions_file.read())
                        if used_versions['_amplimap'] != str(__version__):
                            sys.stderr.write('This analysis was performed with {} {} but this is {} {}!\n\n'.format(__title__, used_versions['_amplimap'], __title__, __version__))
                            sys.stderr.write('Please use the correct version of {} or start a new analysis.\n'.format(__title__))
                            return 1
                        else:
                            sys.stderr.write('{} version checked.\n'.format(__title__))

                #check used config file
                if os.path.isfile(configfile):
                    with open(configfile, 'r') as used_config_file:
                        used_config = yaml.safe_load(used_config_file.read())
                        differences = compare_config_dicts(config, used_config)
                        if len(differences) > 0:
                            sys.stderr.write('config_used.yaml in analysis directory differs from current config.yaml in working directory! Please rename or delete the old analysis directory to restart analysis with the new configuration.\n')
                            sys.stderr.write('Different settings:\n')
                            for diff in differences:
                                sys.stderr.write('\t- {}\n'.format(':'.join(diff)))
                            return 1
                        else:
                            sys.stderr.write('Config files checked.\n')

                #check hashes of input files
                if not args.skip_file_check:
                    used_file_hashes_path = os.path.join(analysis_dir, 'file_hashes.yaml')
                    if os.path.isfile(used_file_hashes_path):
                        with open(used_file_hashes_path, 'r') as used_file_hashes_file:
                            used_file_hashes = yaml.safe_load(used_file_hashes_file.read())

                            from .reader import get_file_hashes
                            for fn, current_hash in get_file_hashes(args.working_directory).items():
                                if used_file_hashes[fn] != current_hash:
                                    sys.stderr.write('File {} seems to have changed since the last run!\n\n'.format(fn))
                                    sys.stderr.write('To ensure consistent results, you should rename or delete the old analysis directory and start a new analysis.\n')
                                    sys.stderr.write('To ignore this error, add the --skip-file-check parameter.\n')
                                    return 1
                            sys.stderr.write('Input files checked.\n')
                else:
                    sys.stderr.write('Warning: Skipping input file check.\n')

        # ensure analysis dir exists now
        os.makedirs(analysis_dir, exist_ok=True)

        # write config to analysis directory, and then use that for snakemake
        with open(configfile, 'w') as f:
            yaml.dump(config, f, default_flow_style=False)

        # set up cluster commands
        cluster_command_nosync = None
        cluster_command_sync = None

        if args.cluster:
            if args.cluster in config['clusters'] and isinstance(config['clusters'][args.cluster], dict):
                if 'command_sync' in config['clusters'][args.cluster]:
                    cluster_command_sync = config['clusters'][args.cluster]['command_sync']
                elif 'command_nosync' in config['clusters'][args.cluster]:
                    cluster_command_nosync = config['clusters'][args.cluster]['command_nosync']
                else:
                    raise Exception('Invalid cluster configuration -- need either command_sync or command_nosync for: {}'.format(args.cluster))
            else:
                raise Exception('Cluster type not found in config: {}'.format(args.cluster))

            sys.stderr.write('Running in cluster mode {} with {} parallel jobs\n'.format(args.cluster, args.njobs))
            sys.stderr.write('cluster_command_nosync={}\n'.format(cluster_command_nosync))
            sys.stderr.write('cluster_command_sync={}\n'.format(cluster_command_sync))

            # make sure cluster log directory exists (this assumed the cluster command is using this as a parameter)
            cluster_logs = os.path.join(args.working_directory, 'cluster_log')
            os.makedirs(cluster_logs, exist_ok=True)
            sys.stderr.write('Will write cluster logs to: {}\n'.format(cluster_logs))
        else:
            sys.stderr.write('Running locally with {} cores\n'.format(args.ncores))

        extra_snakemake_args = {}
        if args.snakemake_args:
            extra_snakemake_args = {
                kv[0]: (True if kv[1].lower() == 'true' else False if kv[1].lower() == 'false' else kv[1])
                for kv in [
                    x.split('=') for x in args.snakemake_args.split(',')
                ]
            }

            sys.stderr.write('Using extra Snakemake arguments: {}\n'.format(str(extra_snakemake_args)))

        success = snakemake.snakemake(
            snakefile = os.path.join(basedir, "Snakefile"),
            configfile = configfile,
            cores = args.ncores,  # ignored if cluster
            nodes = args.njobs,  # ignored if not cluster
            workdir = args.working_directory,
            targets = args.TARGET,
            dryrun = not args.run,
            cluster = cluster_command_nosync,
            cluster_sync = cluster_command_sync,
            jobname = "{}.{{rulename}}.{{jobid}}.sh".format(__title__),
            unlock = args.unlock,
            latency_wait = args.latency_wait,
            **extra_snakemake_args
        )

        sys.stderr.write('\n===============================================\n\n')
        if success:
            if args.unlock:
                sys.stderr.write('Unlocked working directory. Run without --unlock to start.\n')
            elif not args.run:
                sys.stderr.write('{} {} dry run successful. Set --run to run!\n'.format(__title__,  __version__))
            else:
                sys.stderr.write('{} {} finished!\n'.format(__title__, __version__))
            return 0
        else:
            if args.cluster:
                sys.stderr.write('{} {} failed! Please see output above or the cluster log files for details.\n'.format(__title__,  __version__))
                sys.stderr.write('\nFor details on how to find the correct cluster log file for a failed job, see: https://amplimap.readthedocs.io/en/latest/usage.html#cluster-log-files\n')
                sys.stderr.write('You can also try to run amplimap without the cluster parameter to see the error message.\n')
            else:
                sys.stderr.write('{} {} failed! Please see output above for details.\n'.format(__title__,  __version__))

            return 1
    except AmplimapReaderException as e:
        if args.debug:
            import traceback
            traceback.print_exc()
        sys.stderr.write(str(e))
        sys.stderr.write('{} {} failed!\n'.format(__title__, __version__))
        return 2
    except Exception as e:
        if args.debug:
            import traceback
            traceback.print_exc()
        sys.stderr.write('\nERROR: {}\n\n'.format(e))
        sys.stderr.write('{} {} failed!\n'.format(__title__, __version__))
        return 1

if __name__ == '__main__':
    sys.exit(main())



#make sure we can import from package directory
import sys, os
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__)))) 

import unittest
import string, math
import pysam
import amplimap.common
import amplimap.reader
from amplimap.pileup import *

class SimulatedPileupRead():
    pass

class Pileups(unittest.TestCase):
    def test_process_pileup_read(self):
        for mapq in [0, 20, 100]:
            for vtype in ['snp', 'ins', 'del', 'skip']:
                for naive_umi in ['', 'NNNN', 'ACGTACGT']:
                    for add_softclip in [True, False]:
                        for test_validate  in [0, 1, 2, 3, 4, 5]:
                            should_fail = False

                            pr = None #see below
                            is_del = vtype == 'del'
                            is_refskip = vtype == 'skip'
                            indel = -1 if vtype == 'del' else 1 if vtype == 'ins' else 0

                            probes_dict = None
                            reference_name = 'chrZ'
                            reference_pos = 999
                            min_mapq = 20
                            ignore_groups = False
                            use_naive_groups = False
                            validate_probe_targets = False

                            query_position = 0 if vtype == 'snp' else None

                            read_sequence = 'ACGTA' * 10
                            read_qualities = [40] * len(read_sequence)
                            cigarstring = '%dM' % len(read_sequence) if not add_softclip else '5S%dM' % (len(read_sequence) - 5)
                            umi_group = 'group1'
                            probe_id = 'probe1'
                            original_qname = 'read1'
                            qname = amplimap.common.make_extended_read_name(original_qname, probe_id, naive_umi)

                            #test probe validation with different possibilities
                            if test_validate > 0:
                                validate_probe_targets = True
                                if test_validate == 1:
                                    probes_dict = { 'chr': { }, 'target_start0': {}, 'target_end': {} }
                                    should_fail = True
                                elif test_validate == 1:
                                    probes_dict = { 'chr': { probe_id: reference_name }, 'target_start0': { probe_id: reference_pos }, 'target_end': { probe_id: reference_pos+10 } }
                                    should_fail = False
                                elif test_validate == 2:
                                    probes_dict = { 'chr': { probe_id: reference_name }, 'target_start0': { probe_id: reference_pos - 10 }, 'target_end': { probe_id: reference_pos } }
                                    should_fail = False
                                elif test_validate == 3:
                                    probes_dict = { 'chr': { probe_id: reference_name }, 'target_start0': { probe_id: reference_pos - 10 }, 'target_end': { probe_id: reference_pos - 1 } }
                                    should_fail = True
                                elif test_validate == 4:
                                    probes_dict = { 'chr': { probe_id: 'FAIL' }, 'target_start0': { probe_id: reference_pos }, 'target_end': { probe_id: reference_pos+10 } }
                                    should_fail = True
                                elif test_validate == 5:
                                    probes_dict = { 'chr': { probe_id: reference_name, 'OTHER': reference_name },
                                        'target_start0': { probe_id: reference_pos - 10, 'OTHER': reference_pos-10 },
                                        'target_end': { probe_id: reference_pos - 1, 'OTHER': reference_pos+10 } }
                                    should_fail = True

                            pr = SimulatedPileupRead()
                            pr.is_del = is_del
                            pr.is_refskip = is_refskip
                            pr.indel = indel
                            pr.query_position = query_position #position of the read base at the pileup site, 0-based.

                            pr.alignment = pysam.AlignedSegment()

                            #should be returned
                            pr.alignment.qname = qname

                            #will compare to min_mapq
                            pr.alignment.mapping_quality = mapq

                            #will be checking this for 'S'
                            pr.alignment.cigarstring = cigarstring

                            #read sequence bases, including soft clipped bases (None if not present).
                            pr.alignment.query_sequence = read_sequence
                            #Quality scores are returned as a python array of unsigned chars. Note that this is not the ASCII-encoded value typically seen in FASTQ or SAM formatted files. Thus, no offset of 33 needs to be subtracted.
                            pr.alignment.query_qualities = read_qualities

                            tags = {'pr': probe_id, 'UG': umi_group, 'um': naive_umi}
                            pr.alignment.set_tags(tags.items())

                            if mapq < min_mapq or 'N' in naive_umi or add_softclip or should_fail:
                                with self.assertRaises(PileupRowFilterException): 
                                    process_pileup_read(
                                        pr,
                                        probes_dict,
                                        reference_name,
                                        reference_pos,
                                        min_mapq,
                                        ignore_groups,
                                        use_naive_groups,
                                        validate_probe_targets,
                                        debug = False)
                            else:
                                my_call = None
                                my_phred = None
                                if vtype == 'del':
                                    my_call = 'DEL'
                                    my_phred = 40
                                elif vtype == 'skip':
                                    my_call = 'SKIP'
                                    my_phred = 40
                                elif vtype == 'ins':
                                    my_call = 'INS'
                                    my_phred = 40
                                elif vtype == 'snp':
                                    my_call = read_sequence[query_position]
                                    my_phred = read_qualities[query_position]                            

                                expect = (my_call, my_phred, original_qname, umi_group, probe_id, naive_umi)

                                output = process_pileup_read(
                                    pr,
                                    probes_dict,
                                    reference_name,
                                    reference_pos,
                                    min_mapq,
                                    ignore_groups,
                                    use_naive_groups,
                                    validate_probe_targets,
                                    debug = False)

                                self.assertEqual(expect, output)

    def test_process_pileup_read(self):
        pass

    def test_record_read_in_group(self):
        read_calls = {}
        read_calls_expect = {}

        my_call = 'A'
        my_phred = 10
        my_umi = 'ACGTA'
        qname = 'read1'
        group = 'group1'

        #should be added
        read_calls_expect[qname] = [my_call, my_phred, my_umi]
        output = record_read_in_group(read_calls, my_call, my_phred, my_umi, qname)
        self.assertEqual(output, False)
        self.assertEqual(read_calls, read_calls_expect)
        self.assertEqual(len(read_calls), 1)

        #should be added as well
        qname = 'read2'
        read_calls_expect[qname] = [my_call, my_phred, my_umi]
        output = record_read_in_group(read_calls, my_call, my_phred, my_umi, qname)
        self.assertEqual(output, False)
        self.assertEqual(read_calls, read_calls_expect)
        self.assertEqual(len(read_calls), 2)

        #should be ignored
        my_phred = 5
        output = record_read_in_group(read_calls, my_call, my_phred, my_umi, qname)
        self.assertEqual(output, True)
        self.assertEqual(read_calls, read_calls_expect)

        #phred should be changed
        my_phred = 15
        read_calls_expect[qname] = [my_call, my_phred, my_umi]
        output = record_read_in_group(read_calls, my_call, my_phred, my_umi, qname)
        self.assertEqual(output, True)
        self.assertEqual(read_calls, read_calls_expect)

        #call should be changed
        my_call = 'DEL'
        my_phred = 25
        read_calls_expect[qname] = [my_call, my_phred, my_umi]
        output = record_read_in_group(read_calls, my_call, my_phred, my_umi, qname)
        self.assertEqual(output, True)
        self.assertEqual(read_calls, read_calls_expect)

        #call should be N
        my_call = 'INS'
        my_phred = 25
        read_calls_expect[qname] = ['N', my_phred, my_umi]
        output = record_read_in_group(read_calls, my_call, my_phred, my_umi, qname)
        self.assertEqual(output, True)
        self.assertEqual(read_calls, read_calls_expect)

        #call should be set back to the call now, since we got a higher phred
        my_call = 'G'
        my_phred = 30
        read_calls_expect[qname] = [my_call, my_phred, my_umi]
        output = record_read_in_group(read_calls, my_call, my_phred, my_umi, qname)
        self.assertEqual(output, True)
        self.assertEqual(read_calls, read_calls_expect)

        #should raise
        qname = 'read2'
        my_umi = 'NNNN'
        with self.assertRaises(AssertionError):
            record_read_in_group(read_calls, my_call, my_phred, my_umi, qname)

        #check we didn't mess anything up
        self.assertEqual(len(read_calls), 2)

    def test_get_group_consensus(self):
        self.assertEqual( ('DEL', 1, 30), get_group_consensus( [('DEL', 30, None)], min_consensus_count = 1, ignore_groups = False ))
        self.assertEqual( ('SKIP', 1, 30), get_group_consensus( [('SKIP', 30, None)], min_consensus_count = 1, ignore_groups = False ))
        self.assertEqual( ('A', 1, 30), get_group_consensus( [('A', 30, None)], min_consensus_count = 1, ignore_groups = False ))

        self.assertEqual( ('A', 10, 30), get_group_consensus( [('A', 30, None)]*10, min_consensus_count = 2, ignore_groups = False ) )
        self.assertEqual( ('A', 10, 30), get_group_consensus( [('A', 30, None)]*10, min_consensus_count = 10, ignore_groups = False ) )

        self.assertEqual( ('A', 12, 31), get_group_consensus( [('A', 30, None)]*10 + [('G', 40, None), ('A', 31, None), ('A', 20, None)], min_consensus_count = 2, ignore_groups = False ))
        self.assertEqual( ('G', 11, 40), get_group_consensus( [('G', 20, None), ('A', 30, None)]*10 + [('G', 40, None)], min_consensus_count = 2, ignore_groups = False ))
        self.assertEqual( ('G', 11, 40), get_group_consensus( [('G', 20, None), ('A', 30, None)]*10 + [('G', 40, None)], min_consensus_count = 2, ignore_groups = False ))

        #no base call
        with self.assertRaises(PileupGroupFilterException):
            get_group_consensus( [('N', 30, None)], min_consensus_count = 1, ignore_groups = False )
        with self.assertRaises(PileupGroupFilterException):
            get_group_consensus( [('N', 30, None)], min_consensus_count = 1, ignore_groups = True )

        #no consensus (10 and 10)
        with self.assertRaises(PileupGroupFilterException): #10+10
            get_group_consensus( [('G', 20, None), ('A', 30, None)]*10, min_consensus_count = 2, ignore_groups = False )
        with self.assertRaises(PileupGroupFilterException): #15+10+10
            get_group_consensus( [('G', 20, None)]*15 + [('A', 30, None)]*10 + [('C', 30, None)]*10, min_consensus_count = 2, ignore_groups = False )
        #but fine if we ignore groups
        #but note: we just take the first value here, which happens to be G. not ideal, but should be a rare case
        #however, this should never happen in practise because if ignore_groups is true then we should only get one call per group (group=read)
        self.assertEqual( ('G', 10, 20), get_group_consensus( [('G', 20, None), ('A', 30, None)]*10, min_consensus_count = 2, ignore_groups = True ) )

        #min consensus count
        with self.assertRaises(PileupGroupFilterException):
            get_group_consensus( [('A', 30, None)], min_consensus_count = 2, ignore_groups = False )
        #but fine if we ignore groups
        self.assertEqual( ('A', 1, 30), get_group_consensus( [('A', 30, None)], min_consensus_count = 100, ignore_groups = True ))

        #min consensus fraction
        #check default - min 51%
        self.assertEqual( ('A', 2, 30), get_group_consensus( [('A', 25, None), ('G', 20, None), ('A', 30, None)] )) #1+2
        self.assertEqual( ('A', 3, 35), get_group_consensus( [('A', 25, None), ('G', 20, None), ('A', 30, None), ('A', 35, None)] )) #1+3
        with self.assertRaises(PileupGroupFilterException):
            get_group_consensus( [('G', 20, None), ('A', 30, None)] ) #1+1
        with self.assertRaises(PileupGroupFilterException):
            get_group_consensus( [('G', 20, None), ('A', 30, None)]*2 ) #2+2
        self.assertEqual( ('G', 11, 20), get_group_consensus( [('G', 20, None), ('A', 30, None)]*10 + [('G', 20, None)] )) #11+10
        with self.assertRaises(PileupGroupFilterException):
            get_group_consensus( [('G', 20, None), ('A', 30, None)]*10 ) #10+10
        #check with 75%
        self.assertEqual( ('A', 30, 31), get_group_consensus( [('G', 20, None), ('A', 30, None)]*10 + [('A', 31, None)]*20, min_consensus_fraction = 0.75 )) #30+10
        with self.assertRaises(PileupGroupFilterException):
            get_group_consensus( [('G', 20, None), ('A', 30, None)]*10 + [('A', 30, None)]*20, min_consensus_fraction = 0.76 ) #30+10
        #check with 0% (just get the first one for ties)
        self.assertEqual( ('G', 1, 20), get_group_consensus( [('G', 20, None), ('A', 30, None)], min_consensus_fraction = 0.0 )) #1+1
        #...but still the correct one for two
        self.assertEqual( ('A', 2, 30), get_group_consensus( [('A', 25, None), ('G', 20, None), ('A', 30, None)], min_consensus_fraction = 0.0 )) #1+2
        #check with 100%
        self.assertEqual( ('G', 1, 20), get_group_consensus( [('G', 20, None)], min_consensus_fraction = 1.0 )) #1+1

    def umi_groups_check(self, output, groups):
        seen = []
        for group in groups:
            my = None
            for item in group:
                if my is None:
                    my = output[item]
                    assert my not in seen
                    seen.append(my)

                assert output[item] == my, '%s %s %s' % (str(my), str(item), str(output[item]))
        return True

    def test_find_umi_groups(self):
        #easy - just two different UMIs
        self.umi_groups_check( amplimap.common.find_umi_groups({ b'AAAAA': 10, b'GGGGG': 10 }),
            [[b'AAAAA'], [b'GGGGG']])
        #one mismatch, different sizes
        self.umi_groups_check( amplimap.common.find_umi_groups({ b'AAAAA': 10, b'AAAAG': 1, b'GGGGG': 10 }),
            [[b'AAAAA', b'AAAAG'], [b'GGGGG']])
        #one N, different sizes
        self.umi_groups_check( amplimap.common.find_umi_groups({ b'AAAAA': 10, b'NAAAA': 1, b'GGGGG': 10 }),
            [[b'AAAAA', b'NAAAA'], [b'GGGGG']])
        #one mismatch but half size
        self.umi_groups_check( amplimap.common.find_umi_groups({ b'AAAAA': 10, b'AAAAG': 5, b'GGGGG': 10 }),
            [[b'AAAAA', b'AAAAG'], [b'GGGGG']])
        #one mismatch but equal size
        self.umi_groups_check( amplimap.common.find_umi_groups({ b'AAAAA': 10, b'AAAAG': 10, b'GGGGG': 10 }),
            [[b'AAAAA'], [b'AAAAG'], [b'GGGGG']])
        #one N equal size
        self.umi_groups_check( amplimap.common.find_umi_groups({ b'AAAAA': 10, b'NAAAA': 10, b'GGGGG': 10 }),
            [[b'AAAAA'], [b'NAAAA'], [b'GGGGG']]) #note: N is consiered like any other nucleotide, probably correct
        #two mismatches, but one edit to next
        self.umi_groups_check( amplimap.common.find_umi_groups({ b'AAAAA': 10, b'AAAAG': 1, b'TAAAG': 1, b'GGGGG': 10 }),
            [[b'AAAAA', b'AAAAG', b'TAAAG'], [b'GGGGG']])
        #two mismatches, but without the step in-between
        self.umi_groups_check( amplimap.common.find_umi_groups({ b'AAAAA': 10, b'TAAAG': 1, b'GGGGG': 10 }),
            [[b'AAAAA'], [b'TAAAG'], [b'GGGGG']])
        
if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
"""
This module contains methods to parse, merge, annotate and filter variant calls.
"""

import pandas as pd
import numpy as np
import re
import os

# for overlap with targest bed
import collections
import interlap
# for VCF processing
import pysam

import logging
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
sh = logging.StreamHandler()
sh.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
log.addHandler(sh)

#functions for reading input files
from .reader import read_targets, read_sample_info

#datatypes for variant summary tables
variants_dtypes = {
    'Chr': str,
    'Ref': str,
    'Alt': str,
    'Start': int,
    'End': int,
    'GeneDetail.refGene': str,
    'Func.refGene': str,
    'ExonicFunc.refGene': str,
    'genomicSuperDups': str,
    'SIFT_pred': str,
    'Polyphen2_HDIV_pred': str,
    'LRT_pred': str,
    'MutationTaster_pred': str,
    'MutationAssessor_pred': str,
    'FATHMM_pred': str,
    'GERP++_RS': float,
    'phyloP100way_vertebrate': float,
}
variants_na_values = ['', 'NA', '.']

def load_gene_exons(file: str, genes: list) -> dict:
    """Load list of exon chr, strand, start and end locations for each gene in genes."""
    genes = genes.unique()

    #load gene exon information from flat table
    print('Loading exon table from ', file)
    gene_exon_df = pd.read_csv(file, header=None, sep='\t')
    gene_exon_df.rename(columns={2: 'chr', 3: 'strand', 8: 'n_exons', 9: 'exon_starts', 10: 'exon_ends', 12: 'gene'}, inplace=True)
    print('Loaded ', len(gene_exon_df), ' rows.')

    #filter down to only the genes we actually need
    gene_exon_df = gene_exon_df[ gene_exon_df['gene'].isin(genes) ]
    print('Filtered to ', len(gene_exon_df), ' rows for ', len(genes), ' genes.')

    #identify standard chrs
    regex_standard_chr = re.compile(r'^(chr)?([0-9MXY]+)$')

    #make dict of genes to exon locations
    gene_exons = {}
    for gex in gene_exon_df.itertuples():
        try:
            #skip chrY FIXME: do we want this
            if gex.chr == 'chrY':
                continue

            exon_data = {
                'chr': gex.chr,
                'strand': gex.strand,
                'starts': [int(x) for x in gex.exon_starts.split(',') if len(x) > 0],
                'ends': [int(x) for x in gex.exon_ends.split(',') if len(x) > 0],
            }

            #do we already have exons for this gene?
            if gex.gene in gene_exons:
                #is this on the same chr (and not an Alt chr)?
                if gene_exons[gex.gene]['chr'] == exon_data['chr']:
                    #if we are on the same strand let's merge them
                    if gene_exons[gex.gene]['strand'] == exon_data['strand']:
                        n_total = 0
                        n_added = 0
                        for start, end in zip(exon_data['starts'], exon_data['ends']):
                            n_total += 1

                            #try to find existing pair that matches
                            for existing_start, existing_end in zip(gene_exons[gex.gene]['starts'], gene_exons[gex.gene]['ends']):
                                if existing_start == start and existing_end == end:
                                    break
                            #otherwise add it
                            else:
                                n_added += 1
                                gene_exons[gex.gene]['starts'].append(start)
                                gene_exons[gex.gene]['ends'].append(end)

                        #log.debug('Merging exon data from duplicate rows for gene %s - added %d/%d', gex.gene, n_added, n_total)
                    else:
                        # print(gex.gene)
                        # print(gene_exons[gex.gene])
                        # print(exon_data)
                        raise Exception('Exon strand mismatch for {}'.format(gex.gene))
                #if the existing copy is on a standard chr we want to keep it
                elif regex_standard_chr.match(gene_exons[gex.gene]['chr']):
                    #normally this should only happen if we are on an alt chr now
                    if not regex_standard_chr.match(gex.chr):
                        # log.debug('Skipping copy of gene on nonstandard chr for gene exon table: {} @ {}/{}'.format(
                        #     gex.gene, gex.chr, gene_exons[gex.gene]['chr']
                        # ))
                        continue
                    else:
                        raise Exception('Two copies of gene on different standard chrs encounted for gene exon table: {} @ {}/{}'.format(
                            gex.gene, gex.chr, gene_exons[gex.gene]['chr']
                        ))
                #if the existing copy was not on a standard chr we will just overwrite it with this copy
                else:
                    pass
                    # log.debug('Overwriting copy of gene on nonstandard chr for gene exon table: {} @ {}/{}'.format(
                    #     gex.gene, gex.chr, gene_exons[gex.gene]['chr']
                    # ))

            #error checking
            assert len(exon_data['starts']) == len(exon_data['ends'])
            assert len(exon_data['starts']) == gex.n_exons
            for start, end in zip(exon_data['starts'], exon_data['ends']):
                assert end > start, 'Exon end is not after start for {}'.format(gex.gene)

            #finally save the exon data
            gene_exons[gex.gene] = exon_data
        except Exception as e:
            print('Igoring error while processing exon data: {}'.format(
                str(e)))

    #make sure we found exons for every gene
    missing_genes = set(genes) - set(gene_exons.keys())
    if len(missing_genes) > 0:
        print('Failed to find exon location information for: %s' % (','.join(missing_genes)))

    return gene_exons

def find_closest_exon(row: pd.Series, gexs: dict) -> int:
    """Get distance of variant to the closest exon of the gene it has been annotated with."""

    if row['Gene.refGene'] is None:
        return None
    if not row['Func.refGene'] in ['intronic', 'upstream', 'downstream']:
        return None
    if not row['Gene.refGene'] in gexs:
        return None

    gex = gexs[row['Gene.refGene']]
    assert (row['Chr'] == gex['chr']) or (row['Chr'] == 'chr'+gex['chr']) or ('chr'+row['Chr'] == gex['chr'])

    #upstream of start = negative, downstream of end = positive
    start_dist = [row['End'] - boundary_pos for boundary_pos in gex['starts'] if boundary_pos >= row['End']]
    end_dist = [row['Start'] - boundary_pos for boundary_pos in gex['ends'] if boundary_pos <= row['Start']]

    #find smallest abs(value) for each
    closest_start = max(start_dist) if len(start_dist) > 0 else None
    closest_end = min(end_dist) if len(end_dist) > 0 else None

    #figure out whether we are closer to start or end
    if closest_start is None and closest_end is None:
        closest_distance = None
    elif closest_start is None:
        closest_distance = closest_end
    elif closest_end is None:
        closest_distance = closest_start
    else:
        closest_distance = closest_start if abs(closest_start) < closest_end else closest_end

    #if gene strand is negative it's actually the opposite
    if closest_distance is None:
        return None
    elif gex['strand'] == '+':
        return (closest_distance)
    elif gex['strand'] == '-':
        return (closest_distance * -1)
    else:
        raise Exception('Invalid strand')

#see: /hts/data6/smcgowan/hts_scripts/TableAnnovar_to_spreadsheet.pl
def calculate_del_score(merged: pd.DataFrame):
    """
    Add a column DeleteriousScore to dataframe which contains a count of how many tools have assigned this variant a deletious scores.

    Score ranges from 0-6, corresponding to the tools SIFT, Polyphen2, LRT, MutationTaster, GERP++ and phyloP100way_vertebrate.

    Additionally, any stopgain, frameshift or splicing variants are always set to 6.
    """

    merged['DeleteriousScore'] = 0
    if 'SIFT_pred' in merged.columns:
        merged.loc[merged['SIFT_pred'] == 'D', 'DeleteriousScore'] += 1
    if 'Polyphen2_HDIV_pred' in merged.columns:
        merged.loc[merged['Polyphen2_HDIV_pred'] == 'D', 'DeleteriousScore'] += 1
    if 'LRT_pred' in merged.columns:
        merged.loc[merged['LRT_pred'] == 'D', 'DeleteriousScore'] += 1
    if 'MutationTaster_pred' in merged.columns:
        merged.loc[merged['MutationTaster_pred'].isin(['A', 'D']), 'DeleteriousScore'] += 1
    if 'GERP++_RS' in merged.columns:
        assert merged['GERP++_RS'].dtype == float
        merged.loc[merged['GERP++_RS'] >= 5, 'DeleteriousScore'] += 1
    if 'phyloP100way_vertebrate' in merged.columns:
        assert merged['phyloP100way_vertebrate'].dtype == float
        merged.loc[merged['phyloP100way_vertebrate'] > 1, 'DeleteriousScore'] += 1

    #set score = 6 for some special cases
    if 'ExonicFunc.refGene' in merged.columns:
        merged.loc[merged['ExonicFunc.refGene'].notnull() & merged['ExonicFunc.refGene'].str.contains('^stopgain|^frameshift'), 'DeleteriousScore'] = 6
        merged.loc[merged['ExonicFunc.refGene'].notnull() & merged['ExonicFunc.refGene'].str.contains('splic'), 'DeleteriousScore'] = 6
    if 'Func.refGene' in merged.columns:
        merged.loc[merged['Func.refGene'].notnull() & merged['Func.refGene'].str.contains('splic'), 'DeleteriousScore'] = 6

def merge_variants_from_annovar(input, output):
    """Merge individual Annovar CSV files together."""

    merged = None
    for file in input:
        sname = os.path.basename(file).split('.')[0]
        print('Reading', file, ' ( Sample =', sname, ')...')
        try:
            df = pd.read_csv(file, index_col = False, dtype = variants_dtypes, na_values = variants_na_values)

            print('Data shape:', str(df.shape))
            if len(df) > 0:
                df['Sample'] = sname

                if merged is None:
                    merged = df
                else:
                    merged = merged.append(df, ignore_index = True)
            else:
                print ('Ignoring - empty!')
        except pd.errors.EmptyDataError:
            print('No data for', file, ', skipping.')

    if merged is None:
        print('No variants found in any of the samples, creating empty file!')
        open(output[0], 'a').close()
    else:
        print('Merged data shape:', str(merged.shape))
        merged.to_csv(output[0], index = False)

def merge_variants_unannotated(input_vcfs, output_file):
    """Merge unannotated VCF files into a single CSV."""

    merged = None
    for file in input_vcfs:
        sname = os.path.basename(file).split('.')[0]
        fileinfo = os.stat(file)
        print('Reading', file, ' ( Sample =', sname, ') with size', fileinfo.st_size, ' bytes...')

        if fileinfo.st_size == 0:
            print('Empty file, skipping')
        else:
            rows = []
            with pysam.VariantFile(file) as vcf:
                for variant in vcf:
                    assert len(variant.alts) == 1, 'Invalid VCF - multiallelics should have been split!'

                    row = collections.OrderedDict()
                    row['Sample'] = sname
                    row['Chr'] = variant.chrom
                    row['Start'] = variant.start + 1  # pysam is 0-based
                    row['End'] = variant.start + len(variant.ref)
                    row['Ref'] = variant.ref
                    row['Alt'] = ','.join(variant.alts)
                    row['Otherinfo'] = str(variant).strip()
                    rows.append(row)

            if len(rows) > 0:
                print('Found', len(rows), 'rows.')
                df = pd.DataFrame(rows)
                if merged is None:
                    merged = df
                else:
                    merged = merged.append(df, ignore_index = True)
            else:
                print('Ignoring - empty!')

    if merged is None:
        print('No variants found in any of the samples, creating empty file!')
        open(output_file, 'a').close()
    else:
        print('Merged data shape:', str(merged.shape))
        merged.to_csv(output_file, index = False)

def make_summary(input: list, output: list, config: dict, exon_table_path: str = None):
    """Load merged Annovar CSV file (plus targets and sample info), process them and output a new CSV file."""

    #load targets (to add a targets column)
    targets = None
    if len(input['targets']) > 0:
        targets = read_targets(input['targets'], file_type = 'bed', reference_type = 'genome')

    #load sample information table
    sample_info = None
    if len(input['sample_info']) > 0:
        sample_info = read_sample_info(input['sample_info'][0])

    #load merged variant table, process it and write to CSV
    try:
        merged = pd.read_csv(input['merged'], index_col = False, dtype = variants_dtypes, na_values = variants_na_values)
        merged = make_summary_dataframe(
            merged,
            targets,
            sample_info,
            genome_name = config['general']['genome_name'],
            include_gbrowse_links = 'include_gbrowse_links' in config['annotate'] and config['annotate']['include_gbrowse_links'],
            include_exon_distance = 'include_exon_distance' in config['annotate'] and config['annotate']['include_exon_distance'],
            include_score = 'include_score' in config['annotate'] and config['annotate']['include_score'],
            exon_table_path = exon_table_path,
        )
        merged.to_csv(output[0], index = False)
    except pd.errors.EmptyDataError:
        print('No variants found, creating empty file!')
        open(output[0], 'a').close()

def make_summary_dataframe(
        merged: pd.DataFrame,
        targets: pd.DataFrame = None,
        sample_info: pd.DataFrame = None,
        genome_name: str = None,
        include_gbrowse_links: bool = False,
        include_exon_distance: bool = False,
        include_score: bool = False,
        exon_table_path: str = None,
    ) -> pd.DataFrame:
    """Process merged Annovar dataframe (with optional targets and sample_info data frames) into a large summary table."""

    # process targets (to add a targets column)
    target_intervals = None
    if targets is not None:
        target_intervals = collections.defaultdict(interlap.InterLap)
        for target in targets.itertuples():
            target_intervals[target.chr].add( (int(target.start_0), int(target.end), target) ) #note the double parentheses!

    # process sample information table
    sample_info_columns = []
    if sample_info is not None:
        sample_info_columns = list(sample_info.columns)

    # read original data from VCF columns
    vcf = merged['Otherinfo' if 'Otherinfo' in merged.columns else 'Otherinfo1'].apply(lambda x: pd.Series(x.split('\t')))
    vcf.replace('.', np.nan, inplace=True) #replace dots with NAs again
    vcf.columns = ['Chr', 'Pos', 'ID', 'Ref', 'Alt', 'Qual', 'Filter', 'Info', 'Fields', 'SampleData']

    # fix types
    vcf = vcf.astype(
        {
            'Chr': str,
            'Pos': int,
            'ID': str,
            'Ref': str,
            'Alt': str,
            'Qual': float,
            'Filter': str,
            'Info': str,
            'Fields': str,
            'SampleData': str,
        },
        errors = 'raise'
    )

    # make sure chromosome match between VCF data and annotation (they alwyas should)
    assert (
        (merged['Chr'] == vcf['Chr']) |
        (merged['Chr'] == ['chr'+c for c in vcf['Chr']]) |
        (['chr'+c for c in merged['Chr']] == vcf['Chr'])
    ).all(), 'Chromosome names mismatch between Annovar and VCF'

    # set chromosome to the one recorded in VCF, since Annovar may have added 'chr' prefix
    merged['Chr'] = vcf['Chr']

    # We are using the approach from /hts/data6/smcgowan/hts_scripts/TableAnnovar_to_spreadsheet.pl, which only looks for exons
    # for the given gene (as annotated by Annovar). This should work fine, but is a bit inefficient.
    # See here for better, more general solutions: https://www.biostars.org/p/53561/
    # or actually just use the same approach as for targest (interlap.InterLap)
    if include_exon_distance:
        if not 'Gene.refGene' in merged.columns or not 'Func.refGene' in merged.columns:
            log.warning('Variants have to be annotated with the Annovar refGene database to calculate exon distances.')
        else:
            merged['DistanceNearestExon'] = merged.apply(find_closest_exon, axis = 1, gexs = load_gene_exons(
                exon_table_path,
                merged.loc[merged['Func.refGene'].isin(['intronic', 'upstream', 'downstream']), 'Gene.refGene']))

    # get gt info
    vcf_sample = pd.DataFrame([ dict( zip( row.Fields.split(':'), row.SampleData.split(':') ) ) for row in vcf.itertuples() ])
    vcf_sample.replace('.', np.nan, inplace=True) #replace dots with NAs again

    # parse out some fields
    vcf_infos = vcf['Info'].apply(lambda x: pd.Series(dict( [ tuple(pair.split('=')) if '=' in pair else (pair, True) for pair in x.split(';') ] )))
    # handle multiple alt alleles (not usually expected)
    if 'TR' in vcf_infos:
        vcf_infos.loc[vcf_infos['TR'].str.contains(','), 'TR'] = '-1'
        vcf_infos.loc[vcf_infos['NF'].str.contains(','), 'NF'] = '-1'
        vcf_infos.loc[vcf_infos['NR'].str.contains(','), 'NR'] = '-1'
    if 'NR' in vcf_sample:
        vcf_sample.loc[vcf_sample['NR'].str.contains(','), 'NR'] = '-1' #also need to fix this, for assertion below
        vcf_sample.loc[vcf_sample['NV'].str.contains(','), 'NV'] = '-1' #also need to fix this, for assertion below
    # now rename coverage fields and convert to ints
    have_fwd_ref = False
    vcf_infos['Total_Coverage'] = -1
    vcf_infos['Total_Coverage_fwd'] = -1
    vcf_infos['Total_Coverage_rev'] = -1
    vcf_infos['Total_Reads_Alt'] = -1
    vcf_infos['Total_Reads_Alt_fwd'] = -1
    vcf_infos['Total_Reads_Alt_rev'] = -1

    # PLATYPUS will give us TC/TR
    if 'TC' in vcf_infos:
        have_fwd_ref = True
        vcf_infos['Total_Coverage'] = vcf_infos['TC'].astype(int)
        vcf_infos['Total_Coverage_fwd'] = vcf_infos['TCF'].astype(int)
        vcf_infos['Total_Coverage_rev'] = vcf_infos['TCR'].astype(int)
    if 'TR' in vcf_infos:
        vcf_infos['Total_Reads_Alt'] = vcf_infos['TR'].astype(int)
        vcf_infos['Total_Reads_Alt_fwd'] = vcf_infos['NF'].astype(int)
        vcf_infos['Total_Reads_Alt_rev'] = vcf_infos['NR'].astype(int)

    # GATK will give us an AD value
    if 'AD' in vcf_sample:
        have_fwd_ref = False
        vcf_infos['Total_Reads_Ref'] = [int(x.split(',')[0]) for x in vcf_sample['AD']]
        vcf_infos['Total_Reads_Alt'] = [int(x.split(',')[1]) for x in vcf_sample['AD']]
        if 'DP' in vcf_sample:
            #gatk will give us DP
            vcf_infos['Total_Coverage'] = vcf_sample['DP'].astype(int)
        else:
            #mutect2 won't
            vcf_infos['Total_Coverage'] = [int(x.split(',')[0]) + int(x.split(',')[1]) for x in vcf_sample['AD']]

    # some error checking
    if ((vcf_infos['Total_Coverage_fwd'] != -1) & (vcf_infos['Total_Coverage_rev'] != -1)).any():
        assert (vcf_infos['Total_Coverage'] == vcf_infos['Total_Coverage_fwd'] + vcf_infos['Total_Coverage_rev']).all(), 'total coverage mismatch'
    if ((vcf_infos['Total_Reads_Alt_fwd'] != -1) & (vcf_infos['Total_Reads_Alt_rev'] != -1)).any():
        assert ((vcf_infos['Total_Reads_Alt'] == -1) | (vcf_infos['Total_Reads_Alt'] == vcf_infos['Total_Reads_Alt_fwd'] + vcf_infos['Total_Reads_Alt_rev'])).all(), 'variant coverage mismatch'
    if 'NR' in vcf_sample:
        assert ((vcf_infos['Total_Reads_Alt'] == -1) | (vcf_infos['Total_Coverage'] == vcf_sample['NR'].astype(int))).all(), 'coverage match'
    if 'NV' in vcf_sample:
        assert ((vcf_infos['Total_Reads_Alt'] == -1) | (vcf_infos['Total_Reads_Alt'] == vcf_sample['NV'].astype(int))).all(), 'variant coverage match'

    # add info from VCF to table
    merged['Var_Zygosity'] = ['HOM' if gt == '1/1' else 'Het' if gt in ['0/1', '1/0'] else 'REF' if gt in ['0/0'] else '???' for gt in vcf_sample['GT']]
    merged['Var_FailedFilters'] = ['' if (filt == 'PASS' or filt == '.') else filt for filt in vcf['Filter']]
    # mutect2 gives us AF
    if 'AF' in vcf_sample:
        merged['Var_AltFraction'] = [float(x.split(',')[0]) for x in vcf_sample['AF']]
    else:
        merged['Var_AltFraction'] = 1.0 * vcf_infos['Total_Reads_Alt'] / vcf_infos['Total_Coverage']
        merged.loc[vcf_infos['Total_Reads_Alt'] == -1, 'Var_AltFraction'] = None
    merged['Var_TotalCoverage'] = vcf_infos['Total_Coverage']
    if have_fwd_ref:
        merged['Var_Ref_fwd'] = (vcf_infos['Total_Coverage_fwd'] - vcf_infos['Total_Reads_Alt_fwd'])
        merged['Var_Ref_rev'] = (vcf_infos['Total_Coverage_rev'] - vcf_infos['Total_Reads_Alt_rev'])
    merged['Var_Alt'] = vcf_infos['Total_Reads_Alt']
    if have_fwd_ref:
        merged['Var_Alt_fwd'] = vcf_infos['Total_Reads_Alt_fwd']
        merged['Var_Alt_rev'] = vcf_infos['Total_Reads_Alt_rev']
    merged['Var_QualVariant'] = vcf['Qual'].astype(float)
    if 'GQ' in vcf_sample:
        merged['Var_QualSample'] = vcf_sample['GQ'].astype(float)
    else:
        merged['Var_QualSample'] = None
    merged['Comments'] = ''

    # manually add filter status for low coverage
    merged.loc[merged['Var_TotalCoverage'] < 10, 'Var_FailedFilters'] = [filt + ';CovLt10' if len(filt) > 0 else 'CovLt10' for filt in merged.loc[merged['Var_TotalCoverage'] < 10, 'Var_FailedFilters']]

    # calculate the deleteriousness score
    if include_score:
        calculate_del_score(merged)

    # add gbrowse and regionseq links
    if include_gbrowse_links:
        merged['GBrowse'] = ""
        merged['RegionSeq'] = ""
        if genome_name == 'hg19':
            merged['GBrowse'] = [ '=HYPERLINK("https://gbrowse2.molbiol.ox.ac.uk/fgb2/gbrowse/CRANIOFACIAL_GRCh37/?name=%s:%d-%d", "GBrowse")' % (
                row.Chr, row.Start - 25, row.End + 25
            ) for row in merged[['Chr', 'Start', 'End']].itertuples()]
            merged['RegionSeq'] = [ '=HYPERLINK("https://gbrowse2.molbiol.ox.ac.uk/cgi-bin/varSeqRegion.cgi?var_chr=%s&var_posn=%d&ref=%s&var=%s&rm=mode_2", "RegionSeq")' % (
                row.Chr, row.Start, row.Ref, row.Alt
            ) for row in merged[['Chr', 'Start', 'Ref', 'Alt']].itertuples()]
        elif genome_name == 'hg38':
            merged['GBrowse'] = [ '=HYPERLINK("https://gbrowse2.molbiol.ox.ac.uk/fgb2/gbrowse/CRANIOFACIAL_GRCh38/?name=%s:%d-%d", "GBrowse")' % (
                 row.Chr, row.Start - 25, row.End + 25
            ) for row in merged[['Chr', 'Start', 'End']].itertuples()]
        else:
            print('Not generating GBrowse/RegionSeq columns since reference build is neither hg19 nor hg38.')

    if target_intervals is not None:
        found_targets = []
        for row in merged[['Chr', 'Start', 'End']].itertuples():
            # this function may return 0..N items
            found_targets_for_row = target_intervals[row.Chr].find( (row.Start, row.End) )

            # take the first target, if any
            my_target = None
            for target_range in found_targets_for_row:
                my_target = target_range[2].id
                break  # always take the first hit
            found_targets.append(my_target)

        merged['Target'] = found_targets

    # join sample_info
    if sample_info is not None:
        merged = merged.join(sample_info, on = ['Sample', 'Target'], how = 'left')

    # drop useless columns
    merged.drop([c for c in merged.columns if c.startswith('Otherinfo')], axis=1, inplace=True)

    # drop extra populations from ExAC/gnomAD (all columns that start with ExAC/gnomAD and don't end with _ALL)
    for db_prefix in ['ExAC', 'gnomAD_genome', 'gnomAD_exome']:
        merged.drop([column for column in merged.columns if column.startswith(db_prefix) and not column.endswith('_ALL')], axis=1, inplace=True)

    #first columns
    first_cols = []
    if include_score:
        first_cols.append('DeleteriousScore')
    if 'Gene.refGene' in merged.columns:
        first_cols += ['Gene.refGene', 'Func.refGene', 'ExonicFunc.refGene', 'AAChange.refGene']
    if include_exon_distance:
        first_cols.append('DistanceNearestExon')
    first_cols = first_cols + [c for c in merged.columns if c.startswith('Var_') and c != 'Var_Qual']

    #last columns
    last_cols = ['Chr', 'Start', 'End', 'Ref', 'Alt']
    if include_gbrowse_links:
        last_cols.append('GBrowse')
    last_cols += ['Sample'] + sample_info_columns + ['Comments']

    #columns to remove
    ignored_cols = ['GeneDetail.refGene']

    #select and reorder columns
    merged.sort_values(['Sample', 'Chr', 'Start'], inplace=True)
    merged = merged[first_cols + [c for c in merged.columns if not c in first_cols + last_cols + ignored_cols] + last_cols]

    #output
    return merged

def make_summary_condensed(input, output):
    """Make condensed summary table that only contains a subset of columns."""

    #load sample information table
    sample_info_columns = []
    if len(input['sample_info']) > 0:
        sample_info = read_sample_info(input['sample_info'][0])
        sample_info_columns = list(sample_info.columns)

    for do_filter in ['filtered', 'unfiltered']:
        try:
            #load
            merged = pd.read_csv(input['summary'], index_col = False, dtype = variants_dtypes, na_values = variants_na_values)

            #filter
            if do_filter == 'filtered':
                print('Got %d rows with %d passing filters and %d coverage >= 10' % (len(merged), (merged['Var_FailedFilters'].isnull()).sum(), (merged['Var_TotalCoverage'] >= 10).sum()))
                merged = merged[ (merged['Var_FailedFilters'].isnull()) & (merged['Var_TotalCoverage'] >= 10) ]
                print('Got %d rows after filtering' % len(merged))
            else:
                print('Not filtering, got %d rows' % len(merged))

            #find column with dbsnp-id (taking first that starts with avsnp)
            avsnp_columns = [c for c in merged.columns if c.startswith('avsnp')]
            if len(avsnp_columns) > 0:
                merged['ID_dbsnp'] = merged[avsnp_columns[0]]
            #rename column
            merged['Genotype'] = merged['Var_Zygosity']
            #get desired columns
            output_columns = ([] if do_filter == 'filtered' else ['Var_FailedFilters']) + [
                    'Sample', 'Target',
                    'Chr', 'Start', 'Ref', 'Alt', 'Genotype',
                    'ID_dbsnp', 'DeleteriousScore',
                    'Gene.refGene', 'Func.refGene', 'ExonicFunc.refGene', 'AAChange.refGene'
                ] + sample_info_columns
            #make sure the columns actually exist
            output_columns = [c for c in output_columns if c in merged.columns]

            merged = merged[output_columns]

            #output
            merged.to_csv(output[do_filter], index = False)
        except pd.errors.EmptyDataError:
            print('No variants found, creating empty file!')
            open(output[do_filter], 'a').close()

# def make_summary_excel(input, output):
#     """UNTESTED: make Excel table for merged table"""
#     import openpyxl

#     try:
#         merged = pd.read_csv(input[0], index_col = False)

#         #reset header format
#         pd.formats.format.header_style = None

#         #requires openpyxl
#         with pd.ExcelWriter(output[0]) as xlsx:
#             merged.to_excel(xlsx, sheet_name='All (%d)' % len(merged))

#             merged['_sheet'] = 'Unknown'
#             for sheet, regex in {
#                 'ncRNA': 'ncRNA',
#                 'Intronic intergenic': 'intronic|intergenic',
#                 'Splicing': 'splic',
#                 'Up-downstream': 'stream',
#                 'UTR': 'utr'
#             }.items():
#                 merged.loc[merged['Func.refGene'].notnull() & merged['Func.refGene'].str.contains(regex, case=False), '_sheet'] = sheet

#             for sheet, regex in {
#                 'Stops': 'stop',
#                 'Nonsyn': 'nonsynonymous',
#                 'Synon': 'synonymous',
#                 'Fshifts': 'frameshift'
#             }.items():
#                 merged.loc[merged['ExonicFunc.refGene'].notnull() & merged['ExonicFunc.refGene'].str.contains(regex, case=False), '_sheet'] = sheet

#             sheets = ['Nonsyn', 'Stops', 'Fshifts', 'Splicing', 'Synon', 'Intronic intergenic', 'Up-downstream', 'UTR', 'ncRNA', 'Unknown']
#             assert merged['_sheet'].isin(sheets).all()
#             for sheet in sheets:
#                 my_merged = merged[merged['_sheet'] == sheet]
#                 if len(my_merged) > 0:
#                     my_merged.drop('_sheet', axis=1, inplace=True)
#                     my_merged.to_excel(xlsx, sheet_name='%s (%d)' % (sheet, len(my_merged)))

#             #add formatting
#             workbook = xlsx.book
#             for worksheet in workbook:
#                 #http://openpyxl.readthedocs.io/en/default/api/openpyxl.worksheet.worksheet.html
#                 #http://openpyxl.readthedocs.io/en/default/formatting.html
#                 #add colour scale on deleteriousness score
#                 worksheet.conditional_formatting.add('A1:A%d'%worksheet.max_row, openpyxl.formatting.rule.ColorScaleRule(start_color = 'ffffff', end_color = 'aa0000'))
#                 #causes error, not clear why...
#                 pass

#             xlsx.save()
#     except pd.errors.EmptyDataError:
#         print('No variants found, creating empty file!')
#         open(output[0], 'a').close()



#make sure we can import from package directory
import sys, os
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__)))) 

import unittest
import string, math
from amplimap.naive_mapper import *

class NaiveMapper(unittest.TestCase):
    def test_align(self):
        ref = "ACGTAAAAAAAAAAAAAAAG"

        #match
        self.assertEqual(align_and_find_cigar('ACGT', ref), ( 0, [ (0, 4) ] ))
        #match at end
        self.assertEqual(align_and_find_cigar('AAAAG', ref), ( 15, [ (0, 5) ] ))
        #self.assertEqual(align_and_find_cigar('AAAAG', ref, reverse = True), ( 15, [ (0, 5) ] ))
        #match with deletion
        self.assertEqual(align_and_find_cigar('AGTAAA', ref), ( 0, [ (0, 1), (2, 1), (0, 5) ] ))
        #match with insertion (but actually ends up getting matched as A skip and C mismatch)
        self.assertEqual(align_and_find_cigar('ACCGT', ref), ( 0, [(4, 1), (0, 4)] ))
        #proper match with insertion 
        self.assertEqual(align_and_find_cigar('ACGGTAAA', ref), ( 0, [ (0, 2), (1, 1), (0, 5) ] ))
        #soft clip start
        self.assertEqual(align_and_find_cigar('GGGACGT', ref), ( 0, [ (4, 3), (0, 4) ] ))
        #soft clip start and insert
        self.assertEqual(align_and_find_cigar('GGGACGGTA', ref), ( 0, [ (4, 3), (0, 2), (1, 1), (0, 3) ] ))
        #start offset
        self.assertEqual(align_and_find_cigar('TAAAAA', ref), ( 3, [ (0, 6) ] ))
        #self.assertEqual(align_and_find_cigar('TAAAAA', ref, reverse = True), ( 3, [ (0, 6) ] ))
        #start offset and change
        self.assertEqual(align_and_find_cigar('TAAAGAA', ref), ( 3, [ (0, 7) ] ))
        #start offset and insert (need at least five to be better than mismatch!)
        self.assertEqual(align_and_find_cigar('TAAAGGGGGAA', ref), ( 3, [ (0, 4), (1, 5), (0, 2) ] ))


if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
"""
This is a custom read mapper for amplicon data.
It simply places the reads at the location they were expected to come from,
based on the target coordinates of their associated probe.
It will perform an aligment within the target region to handle small differences
but will not correctly handle chimeric off-target reads or errors.
"""

import sys, os, re, time, collections

import logging
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
sh = logging.StreamHandler()
sh.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
log.addHandler(sh)

#for bam creation
import pysam
#for getting ref chrs
import pyfaidx
#for fastq reading
import Bio.SeqIO
import Bio.pairwise2
import gzip

#amplimap imports
from .common import parse_extended_read_name
from .reader import read_new_probe_design
from .version import __title__, __version__

class AmplimapNoAlignment(Exception):
    pass

def align_and_find_cigar(read, ref, debug = False, reverse = False) -> tuple:
    """
    Align read and reference sequence using parwise global alignment and return start offset and CIGAR ops.
    """
    assert not reverse, 'not in use'

    #we always want to be left-aligned with the proper seq (although it's not clear whether this actually makes a diff)
    if reverse:
        read = read[::-1]
        ref = ref[::-1]
    alignments = Bio.pairwise2.align.globalms(read, ref, 2, -1, -3, -.5, one_alignment_only = True, penalize_end_gaps = False)
    if len(alignments) == 0:
        raise AmplimapNoAlignment()

    alignment = alignments[0]
    #now we need to reverse these again, so that the cigar string (which is in genomic orientation) makes sense
    if reverse:
        alignment = (alignment[0][::-1], alignment[1][::-1])
    return find_cigar_for_alignment(len(read), alignment, debug)

def find_cigar_for_alignment(read_len, alignment, debug) -> tuple:
    """
    Generate tuple of start offset and CIGAR operations for given alignment.
    """
    cigartuples = []
    cigar_ref_consumed = 0
    cigar_read_consumed = 0
    cigar_read_start_offset = 0

    if debug:
        print(alignment)

    for char_read, char_ref in zip(alignment[0], alignment[1]):
        char_cigar = None
        if char_read == '-' and char_ref == '-':
            raise Exception('both gaps?!')
        elif char_read == '-':
            #if we already aligned some bases then this is a deletion
            if cigar_read_consumed > 0:
                char_cigar = 2 #DEL
            #otherwise we have a start offset
            else:
                cigar_read_start_offset += 1
            cigar_ref_consumed += 1
        elif char_ref == '-':
            #if we already started the read this is an insertion, otherwise we soft-clip
            if cigar_ref_consumed > 0:
                char_cigar = 1 #INS
            else:
                char_cigar = 4 #SOFTCLIP
            cigar_read_consumed += 1
        else:
            char_cigar = 0 #MATCH
            cigar_read_consumed += 1
            cigar_ref_consumed += 1

        #this may be None if we have start offset
        if char_cigar is not None:
            #start new tuple or increase counter for current tuple
            if len(cigartuples) == 0 or cigartuples[-1][0] != char_cigar:
                cigartuples.append( [char_cigar, 1] )
            else:
                cigartuples[-1][1] += 1

        # if debug:
        #     print(char_read, cigar_read_consumed, char_ref, cigar_ref_consumed, char_cigar, cigar_read_start_offset)

        #cigar should not get longer than read
        if cigar_read_consumed == read_len:
            break

    if debug:
        print(cigar_read_start_offset, cigartuples, sum( [x[1] for x in cigartuples if x[0] in [0, 1, 4]] ))
        print(read_len)

    #ensure that the last operation isn't an insertion (it should just be soft-clipped)
    if cigartuples[-1][0] == 1:
        cigartuples[-1][0] = 4

    #make sure we consumed the entire read
    assert cigar_read_consumed == read_len
    assert sum( [x[1] for x in cigartuples if x[0] in [0, 1, 4]] ) == read_len

    #make these into actual tuples
    cigartuples = [ tuple(x) for x in cigartuples ]

    return cigar_read_start_offset, cigartuples

def create_bam(
    sample,
    files_in1,
    files_in2,
    ref_fasta,
    probes_dict,
    output,
    has_trimmed_primers = True,
    debug = False
):
    """
    Create a BAM file with reads placed at their expected locations, adjusted through pairwise alignment to the target sequences.

    This will give reasonable results as long as probes capture the exact target sequences, but will generate
    alignments with many mismatches if there are any discrepancies.
    """
    assert len(files_in1) == len(files_in2)

    tStart = time.time()
    counters = collections.Counter()

    ref_idx = pyfaidx.Faidx(ref_fasta, rebuild=False)
    bam_header = {
        'HD': { 'VN': '1.0' },
        'SQ': [ {'LN': record.rlen, 'SN': name} for name, record in ref_idx.index.items() ],
        'RG': [ { 'ID': sample, 'SM': sample } ],
        'PG': [ { 'ID': __title__, 'PN': __title__, 'VN': __version__ } ],
    }

    chr_indices = { chrom: index for index, chrom in enumerate(ref_idx.index.keys()) }

    with pysam.AlignmentFile(output, "wb", header = bam_header) as pairedreads:
        for ixfile in range(len(files_in1)):
            file1 = files_in1[ixfile]
            file2 = files_in2[ixfile]
            log.info('Processing %s and %s (#%d)', file1, file2, ixfile)

            counters['files'] += 1
            opener = gzip.open if file1.endswith('.gz') else open
            with opener(file1, 'rt') as hdl1, opener(file2, 'rt') as hdl2:
                for read_pair in zip( Bio.SeqIO.QualityIO.FastqGeneralIterator(hdl1),
                        Bio.SeqIO.QualityIO.FastqGeneralIterator(hdl2) ):
                    counters['pairs_total'] += 1
                    if counters['pairs_total'] % 50000 == 0:
                        log.info("processed %d pairs - %.f sec elapsed, %.4f sec/pair, %.1f pairs/sec",
                            counters['pairs_total'],
                            time.time() - tStart,
                            (time.time() - tStart) / counters['pairs_total'],
                            counters['pairs_total'] / (time.time() - tStart))

                    if debug and counters['pairs_total'] > 10:
                        print('DEBUG - stopping after ', counters['pairs_total'])
                        break

                    #extract and parse read name
                    read_names_original = [
                        read_pair[0][0].split('\t')[0],
                        read_pair[1][0].split('\t')[0],
                    ]
                    assert len(read_names_original[0]) > 0
                    assert read_names_original[0] == read_names_original[1]
                    read_name, read_probe, read_umi = parse_extended_read_name(read_names_original[0])

                    probe_chr = probes_dict['chr'][read_probe]
                    if not probe_chr in chr_indices:
                        raise Exception('Probe {} is associated with chromosome {}, but this entry does not exist in the reference fasta file!'.format(
                            read_probe, probe_chr
                        ))
                    probe_chr_index = chr_indices[probe_chr]

                    read_lens = [ len(read_pair[read_number][1]) for read_number in range(2) ]

                    #untested: if we haven't trimmed off the primers then we need to start aligning from the primer start location!
                    if has_trimmed_primers:
                        probe_start = int(probes_dict['target_start_0'][read_probe] + 1)
                        probe_end = int(probes_dict['target_end'][read_probe])
                    else:
                        probe_start = int(probes_dict['probe_start_0'][read_probe] + 1)
                        probe_end = int(probes_dict['probe_end'][read_probe])

                    if probes_dict['strand'][read_probe] == '+':
                        read_starts = [ probe_start, probe_end - read_lens[1] + 1 ]
                        read_reverse = [ False, True ]
                    elif probes_dict['strand'][read_probe] == '-':
                        read_starts = [ probe_end - read_lens[0] + 1, probe_start ]
                        read_reverse = [ True, False ]
                    else: raise Exception('Unexpected strand for probe {}'.format(read_probe))

                    #NOTE: this SHOULD BE one-based based on documentation
                    #but actually seem to be ZERO-based -- at least the sequence we get for PRRX1-Ex1
                    #starts CGGA but should start GGA; ends TTC but should end TTCT if we just use probe_start and probe_end
                    #they are always in genomic sense
                    probe_target_sequence = str(ref_idx.fetch(probe_chr, probe_start, probe_end)).upper()

                    #sanity check that we got the right sequence
                    if has_trimmed_primers:
                        assert len(probe_target_sequence) == probes_dict['target_length'][read_probe]
                    else:
                        assert len(probe_target_sequence) == probes_dict['capture_size'][read_probe]

                    if debug:
                        print(read_name, read_probe, read_umi)
                        print(probe_chr, probe_chr_index, probe_start, probe_end)
                        print(read_starts)
                        print(read_reverse)
                        print(probe_target_sequence)
                        print(read_pair)

                    try:
                        #pre-process alignments to make sure the mate starts are actually correct
                        read_cigars = []
                        read_sequences = []
                        read_tags_for_pysam = []
                        for read_number in range(2):
                            #copy over our custom tags from FASTQ file
                            read_tags = [ tag.split(':') for tag in read_pair[read_number][0].split('\t')[1:] ]
                            read_tags_for_pysam.append (
                                [("RG", sample, "Z")] + [
                                    (tag_name, int(tag_value) if tag_type == 'i' else tag_value, tag_type) for tag_name, tag_type, tag_value in read_tags
                                ]
                            )
                            if debug:
                                print(read_tags)
                                print(read_tags_for_pysam[read_number])

                            #figure out sequence
                            read_sequence = str(Bio.Seq.Seq(read_pair[read_number][1]).reverse_complement()) if read_reverse[read_number] else read_pair[read_number][1]
                            read_sequences.append(read_sequence)

                            #align read to target sequence -- note both of these are in genomic sense!
                            try:
                                cigar_read_start_offset, cigartuples = align_and_find_cigar(read_sequence, probe_target_sequence)
                                read_tags_for_pysam[read_number].append( ("so", cigar_read_start_offset, 'i') )

                                #remember cigar
                                read_cigars.append(cigartuples)
                                #adjust start -- need to use zero-based coords here but probe_start is 1-based
                                read_starts[read_number] = probe_start - 1 + cigar_read_start_offset
                            except AssertionError:
                                cigar_read_start_offset, cigartuples = align_and_find_cigar(read_sequence, probe_target_sequence,
                                    debug = True)
                                raise

                        if debug:
                            print(read_cigars)
                            print(read_starts)

                        for read_number in range(2):

                            #create aligned segment
                            a = pysam.AlignedSegment()
                            a.mapping_quality = 255 #always best quality
                            a.query_name = read_names_original[0]
                            a.query_sequence = read_sequences[read_number]
                            a.query_qualities = pysam.qualitystring_to_array(read_pair[read_number][2][::-1] if read_reverse[read_number] else read_pair[read_number][2])
                            a.set_tags( read_tags_for_pysam[read_number] )
                            a.cigartuples = read_cigars[read_number]

                            a.reference_id = probe_chr_index
                            a.reference_start = read_starts[read_number]
                            a.next_reference_id = probe_chr_index
                            a.next_reference_start = read_starts[1 - read_number]
                            #a.template_length = read_lens[read_number]

                            a.is_paired = True
                            a.is_proper_pair = True
                            a.is_read1 = read_number == 0
                            a.is_read2 = read_number == 1
                            a.is_reverse = read_reverse[read_number]
                            a.mate_is_reverse = read_reverse[1-read_number]

                            if debug:
                                print(a)
                            pairedreads.write(a)
                            if debug:
                                break
                    #normally we always get an alignment, but apparently sometimes we don't?
                    except AmplimapNoAlignment:
                        counters['no_alignment'] += 1
                        pass


    log.info('%s done - %d pairs in total, %d without alignment', sample, counters['pairs_total'], counters['no_alignment'] )

    log.info("BAM file created: %s", output)


# -*- coding: utf-8 -*-
"""
This module contains methods for generating statistics about the alignments generated by amplimap.
This includes the number of reads and read families per probe, per sample as well as data on off-target
alignments.
"""

#python 3 compat
#http://python-future.org/compatible_idioms.html
from __future__ import print_function

import sys
import os
import re
import pprint

from .common import parse_extended_read_name, find_umi_groups

"""
NOTES:
The BAM contains alignments of reads pairs. We can have these situations:
- Multiple possible alignments with different properties, each can be one of:
    - Alignment of both mates (both have is_unmapped & mate_is_unmapped set to False)
    - Alignment of one mate, other unmapped (one has is_unmapped set, one has mate_is_unmapped set)

- Or: Both unmapped (both have is_unmapped and mate_is_unmapped set) - if this is the case, it should be the only entry for this read pair

APPROACH:
- The BAM is coordinate sorted, so we need to read through it and cache, keeping track of mate positions to find the right mates for each pair.
- For each new mate, we can either:
    - If self is mapped:
        - If mate is mapped:
            - Already have the mate -> process and remove mate from cache
            - Not have the mate -> add to cache and wait for mate
            - NB TODO: actually this will fail with bowtie2 because one read alignment can have multiple mate alignments.
                - should keep all reads in cache, ideally forever or at least per chr in case there are more alignments
                - at the moment we are just ignoring these and outputting these to the log as orphaned reads at the end
        - If not:
            - Count as alignment with unmapped mate [but maybe only if no better alignment]
    - If not:
        - If mate1, count unmapped pair, else ignore (should never see 2x)
- By the end of the file we should have resolved all mates. Anything left is suspicous!

But, question how we deal with multiple alignments per pair. Actually we are only interested in the top aligment, no need to record every single bad secondary alignment.

"""

import logging
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
sh = logging.StreamHandler()
sh.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
log.addHandler(sh)
import time

#use pandas to read the CSV file and write output files
import pandas as pd
#for defaultdict + sorting
import collections
import operator

#use python's argumentparser to support command line parameters like --probe=62
import argparse

#for pileup
import pysam
#for getting ref base
import pyfaidx
#for debugging umi distance
import distance

#import reader script for reading probe design
from .reader import read_new_probe_design

def aggregate(folder):
    log.info('Aggregating tables in %s', folder)

    agg = None
    files = [file for file in os.listdir(folder) if file.endswith('.stats_alignment.csv')]
    files.sort()
    for file in files:
        sample = file
        sample = sample.replace('.stats_alignment.csv', '')
        sample = sample.replace('_L001_R1_001', '')
        sample = sample.replace('__MIP_TRIMMED_', '')

        try:
            df = pd.read_csv(os.path.join(folder, file))
            log.info('Found %d rows in %s -> %s', len(df), file, sample)

            df['sample'] = sample

            if agg is None:
                agg = df
            else:
                agg = agg.append(df)
        except pd.errors.EmptyDataError:
            log.exception("No data found: %s", file)

    assert agg is not None and len(agg) > 0, \
        '\n\nABORTED: Did not find any valid read alignments for any sample. Please check configuration, probe design table and samples!\n\n'

    agg.sort_values(['sample', 'probe'], inplace=True)
    agg = agg[ ['sample', 'probe'] + [c for c in agg.columns if not c in ['sample', 'probe']] ]

    log.info('Found %d rows total, %d columns', len(agg), agg.shape[1])

    outname = os.path.join(folder, 'stats_alignment.csv')
    agg.to_csv(outname, index=False)
    print(outname)

def process_file(
    probes_path: str,
    input_path: str,
    output_path: str,
    min_mapq: int,
    min_consensus_count: int,
    include_primers: bool,
    use_naive_groups: bool,
    ignore_groups: bool,
    ignore_group_mismatch: bool = False,
    debug: bool = False,
    targets_path: str = None,
):
    """
    Loop through coordinate-sorted BAM file and collect alignment statistics.

    Args:
        probes_path (str): Path to probes design CSV file
        input_path (str): Path to input BAM file
        output_path (str): Prefix of output CSV file ('.stats_alignment.csv' will be appended)

    Output file columns:

    - read_pairs_total: Total number of read pairs (one pair may have multiple alignments)
    - alignments_total: Total number of read pair alignments (one pair may have multiple alignments)
    - alignments_good: Total number of alignments that were on target (coordinates match target coordinates) and fully covered the target region
    - umis_good: Total number of UMI groups (read families) that were on target and fully covered the target region
    - umis_good_coverage_ge_NN: Total number of UMI groups (read families) that were on target, fully covered the target region and also had at least NN supporting read pairs
    - alignments_partial: Total number of alignments that were on target but did not fully cover the target region
    - alignments_off_target: Total number of alignments with coordinates different from the target coordinates
    - alignments_unmapped_single: Total number of alignments where one mate was unmapped
    - alignments_unmapped_both: Total number of alignments where both mates were unmapped
    - alignments_flagged: Total number of alignments where at least one mate was flagged as QC fail or supplementary alignment
    - alignments_invalid_pair: Total number of alignments where mates were on different chromosomes or not in the expected orientation (one forward, one reverse)
    - alignments_mapq_lt_NN: Total number of alignments with at least one mate with mapping quality under NN
    - multimapping_alignments: Total number of read pairs that had multiple possible alignments
    - offtarget_locations: Locations and counts for the three most common off-target alignment locations

    Will only process a read pair once both alignments have been found, which means that
    the first mate has to be kept in memory until the second mate has been reached.
    This may lead to high memory usage if mates are far away and there are many alignments
    but seems to work fine in practice.

    However, reads with multiple alternative mate alignments may not be handled
    properly since an alignment is only kept in memory until the first matching mate has been found.
    These will be reported as "orphaned mates" at the end of the run.
    """
    assert input_path is not None, 'input file path missing'
    assert output_path is not None, 'output file path missing'

    #if we did not get a design path we can only look at on/off target
    if probes_path is None or len(probes_path) == 0:
        assert target_path is not None, 'target file path is missing'
        #TODO

    design = read_new_probe_design(probes_path)

    if include_primers:
        probe_column_start = 'probe_start'
        probe_column_end = 'probe_end'
    else:
        probe_column_start = 'target_start'
        probe_column_end = 'target_end'
    log.info('Using columns %s / %s to determine target regions', probe_column_start, probe_column_end)

    log.info('Processing file %s', input_path)
    if not os.path.isfile(input_path):
        sys.exit('input does not exist: %s' % input_path)

    with pysam.AlignmentFile(input_path, "rb") as samfile:
        t_shown = t_start = time.time()

        reads = collections.defaultdict(dict)
        reads_processed = set()
        reads_umi_checked = set()

        n_rows = 0
        n_pairs = collections.defaultdict(int)
        n_alignments = collections.defaultdict(int)
        n_alignments_good = collections.defaultdict(int)
        n_alignments_partial = collections.defaultdict(int)
        n_alignments_off_target = collections.defaultdict(int)
        n_alignments_unmapped_single = collections.defaultdict(int)
        n_alignments_unmapped_both = collections.defaultdict(int)
        n_alignments_flagged = collections.defaultdict(int)
        n_alignments_invalid = collections.defaultdict(int)
        n_alignments_low_mapq = collections.defaultdict(int)
        n_alignments_multimappers = collections.defaultdict(int)

        probe_umi_alignments = collections.defaultdict(collections.Counter)
        n_umis_good = collections.defaultdict(int)
        n_umis_good_coverage_ge_MIN = collections.defaultdict(int)
        probe_offtarget_locations = collections.defaultdict(collections.Counter)

        #make sure we have zeros for each probe id
        for pname in design['id']:
            for counter in [
                n_pairs,
                n_alignments,
                n_alignments_good,
                n_alignments_partial,
                n_alignments_off_target,
                n_alignments_unmapped_single,
                n_alignments_unmapped_both,
                n_alignments_flagged,
                n_alignments_invalid,
                n_alignments_low_mapq,
                n_alignments_multimappers,

                n_umis_good,
                n_umis_good_coverage_ge_MIN,
            ]:
                counter[pname] = 0

        iter = samfile.fetch(until_eof = True)
        for x in iter:
            qname = x.query_name
            assert qname is not None
            rindex = 1 if x.is_read1 else 2 if x.is_read2 else None
            assert rindex is not None
            other_rindex = 3 - rindex

            #supplementary alignments mess up the assumption that each mate has one other mate
            #so we ignore them for now
            if x.is_supplementary:
                #log.warn('Ignorning supplementary alignment for %s', x.query_name)
                continue

            #only count rows after this to fulfil assumption later
            n_rows += 1

            #parse info from read name (should be bowtie2 compatible)
            read_name, read_probe, read_umi = parse_extended_read_name(qname)

            rdata = {}
            rdata['index'] = rindex
            rdata['mapq'] = x.mapping_quality
            rdata['has_bad_flags'] = (x.is_qcfail) or \
                (x.is_supplementary)
                #(x.is_secondary) or \ #TODO: removed this to allow secondary hits from bowtie2, not sure if good idea or not. this column is often (always?) zero for bwa runs though...
            rdata['unmapped'] = x.is_unmapped
            if x.is_unmapped:
                rdata['chr'] = None
                rdata['start'] = None
                rdata['end'] = None
                rdata['reverse'] = None
            else:
                rdata['chr'] = x.reference_name
                rdata['start'] = x.reference_start + 1
                rdata['end'] = x.reference_end
                rdata['reverse'] = x.is_reverse
            rdata['probe'] = read_probe
            rdata['umi'] = read_umi

            pname = read_probe
            if debug:
                print('>', qname, '-', rindex, '(', rdata['probe'], '/', rdata['chr'], ':', rdata['start'], '-', rdata['end'], ')', 'mate mapped =', not x.mate_is_unmapped)

            #keep track of whether we processed this alignment pair now
            unprocessed = False

            #are we unmapped
            if x.is_unmapped:
                if debug:
                    print('is unmapped -- mate_is_unmapped =', x.mate_is_unmapped, ', read1 =', x.is_read1)

                #is our mate also unmapped, and this is read1
                if x.mate_is_unmapped and x.is_read1:
                    n_alignments_unmapped_both[pname] += 1
                else:
                    #if this is read2 we already counted read1 (or will count)
                    #if mate is not unmapped we will count this when we read the mate
                    unprocessed = True
            #are we mapped but our mate is unmapped?
            elif x.mate_is_unmapped:
                if debug:
                    print('is mapped but mate_is_unmapped ==', x.mate_is_unmapped)

                n_alignments_unmapped_single[pname] += 1
            #mapped pair
            else:
                #figure out where the mate should be
                mate_reference_id = x.next_reference_id
                mate_start = x.next_reference_start + 1 #one-indexed to agree with above

                #did we already see the paired mate?
                paired_mate_number = None
                if other_rindex in reads[qname]:
                    for other_mate_number, other_mate_alignment in enumerate(reads[qname][other_rindex]):
                        #is this the correct alignment that we are mated with?
                        if (mate_reference_id == 0 or mate_reference_id == other_mate_alignment['reference_id']) and mate_start == other_mate_alignment['start']:
                            paired_mate_number = other_mate_number
                            break

                if debug:
                    print('mate @ #', mate_reference_id, ':', mate_start, ' -> paired_mate_number =', paired_mate_number)

                if paired_mate_number is None:
                    #for debugging
                    rdata['qname'] = qname
                    rdata['mate_reference_id'] = mate_reference_id
                    rdata['mate_start'] = mate_start

                    #we will be matching using the id, not the chr name
                    rdata['reference_id'] = x.reference_id

                    #keep this read around until we find the mate
                    if not rindex in reads[qname]:
                        reads[qname][rindex] = list()
                    reads[qname][rindex].append(rdata)
                    unprocessed = True
                else:
                    #process the pair
                    rd = {}
                    rd[rindex] = rdata
                    rd[other_rindex] = reads[qname][other_rindex][paired_mate_number]

                    if debug:
                        print('Processing pair:')
                        print(rd)

                    #make sure this is a good pair
                    is_valid_pair = True
                    is_valid_pair = is_valid_pair and rd[1]['reverse'] != rd[2]['reverse']
                    is_valid_pair = is_valid_pair and rd[1]['chr'] == rd[2]['chr']
                    if debug:
                        print('is_valid_pair =', is_valid_pair)

                    #sanity checks
                    if rd[1]['probe'] != rd[2]['probe']:
                        log.error('Probe mismatch for %s', qname)
                        print(rd)
                        assert False

                    #this should not happen anymore, since we catch these abvoe
                    assert not (rd[1]['unmapped'] or rd[2]['unmapped'])

                    #check probe data
                    probe_data = design.loc[pname, :]
                    covers_entire_probe = False
                    matches_probe = rd[1]['chr'] == probe_data['chr']
                    if matches_probe:
                        if rd[1]['reverse']:
                            matches_probe = matches_probe and rd[2]['start'] == probe_data[probe_column_start]
                            matches_probe = matches_probe and rd[1]['end'] == probe_data[probe_column_end]
                            covers_entire_probe = rd[2]['end'] >= rd[1]['start']
                        else:
                            matches_probe = matches_probe and rd[1]['start'] == probe_data[probe_column_start]
                            matches_probe = matches_probe and rd[2]['end'] == probe_data[probe_column_end]
                            covers_entire_probe = rd[1]['end'] >= rd[2]['start']
                    if debug:
                        print('matches_probe =', matches_probe)

                    #decide which bin to put this in
                    if not is_valid_pair:
                        n_alignments_invalid[pname] += 1
                    elif rd[1]['has_bad_flags'] or rd[2]['has_bad_flags']:
                        n_alignments_flagged[pname] += 1
                    elif rd[1]['mapq'] < min_mapq or rd[2]['mapq'] < min_mapq:
                        n_alignments_low_mapq[pname] += 1
                    elif matches_probe:
                        if not covers_entire_probe:
                            n_alignments_partial[pname] += 1
                        else:
                            n_alignments_good[pname] += 1

                            assert rd[1]['umi'] == rd[2]['umi']
                            probe_umi_alignments[pname][ rd[1]['umi'].encode('utf-8') ] += 1
                    else:
                        #otherwise we had a good mapping, but not in the expected location
                        n_alignments_off_target[pname] += 1

                        start_name = '%s:%d-%d' % (rd[1]['chr'], rd[1]['start'], rd[2]['end'])
                        if rd[1]['reverse']:
                            start_name = '%s:%d-%d' % (rd[1]['chr'], rd[2]['start'], rd[1]['end'])
                        probe_offtarget_locations[pname][start_name] += 1

                    if debug and len(reads_processed) > 5:
                        log.warn('Breaking for debug')
                        break

                    #remove paired mate from memory
                    del reads[qname][other_rindex][paired_mate_number]
                    #also remove the containing lists
                    if len(reads[qname][other_rindex]) == 0:
                        del reads[qname][other_rindex]
                    if len(reads[qname]) == 0:
                        del reads[qname]

            #keep track of processed read pairs
            if not unprocessed:
                n_alignments[pname] += 1

                #remember we processed this one
                if qname in reads_processed:
                    n_alignments_multimappers[pname] += 1
                else:
                    n_pairs[pname] += 1
                    reads_processed.add(qname)

            #print some stats
            t_now = time.time()
            if n_rows > 0 and (n_rows % 10000 == 0 or (t_now - t_shown > 60)):
                t_shown = t_now
                log.info("processed %d rows - %.f sec elapsed, %.4f sec/pair, %.1f row/sec",
                    n_rows,
                    t_now - t_start,
                    (t_now - t_start) / n_rows,
                    n_rows / (t_now - t_start))

        #summary stats
        log.info('processed %d alignment rows total.', n_rows)
        log.info('processed %d reads total.', len(reads_processed))

        #check if we have any reads left over?!
        orphan_reads = [r for rs in reads.values() for r in rs.values()]
        if len(orphan_reads) > 0:
            #TODO: this may happen if a single aligned read is paired with multiple alignments of its mate.
            #we will remember the first mate, process the pair once we get the second mate and then forget both
            #next time we see a new alignment for one of the mates we can never find the other alignment, since we have already deleted it
            #I don't think there is an easy way to resolve this unfortunately.
            #Options would be to keep _everything_ in memory for each contig (and lose a few), or to use name-sorted BAM files,
            #or to see if bowtie2 gives us some idea of how many mates there are for a given read, so we know which ones to keep around for how long (reference counting)?
            log.info('Found %d orphaned read mates. First 10:', len(orphan_reads))
            pprint.pprint(orphan_reads[0:10])

            #most of these will be on different references (in our test case all are)
            orphan_reads_same_ref = [orphan_read for orphan_read_list in orphan_reads for orphan_read in orphan_read_list if orphan_read['mate_reference_id'] == orphan_read['reference_id']]
            if len(orphan_reads_same_ref) > 0:
                log.warn('Of these, %d orphaned read mates with same ref id. First 10:', len(orphan_reads_same_ref))
                pprint.pprint(orphan_reads_same_ref[0:10])
            else:
                log.info('However, all of these involve pairs crossing chromosomes.')

        #at the end, process the umi stats
        for (pname, umi_raw_counts) in probe_umi_alignments.items():
            #use umi_tools to group raw UMIs together
            umi_to_group = find_umi_groups(umi_raw_counts)

            # import pprint
            # umi_group_debug = collections.defaultdict(list)
            # for umi, group in umi_to_group.items():
            #     umi_group_debug[group].append('%s=%d' % (umi, umi_raw_counts[umi]))
            # print(pname)
            # pprint.pprint(umi_group_debug)

            #now do a new count
            umi_group_counts = collections.Counter()
            for umi, group in umi_to_group.items():
                #we want the number of reads supporting the umi group here, which is
                #the sum of the numbers of reads supporting each grouped umi
                umi_group_counts[group] += umi_raw_counts[umi]

            #now use these counts
            for _, count in umi_group_counts.items():
                n_umis_good[pname] += 1
                if count >= min_consensus_count:
                    n_umis_good_coverage_ge_MIN[pname] += 1

        #and find the top offtarget locations for each probe
        top_offtarget_locations = dict(
            (
                ( pname, ';'.join(['%s=%d' % (location, count) for location, count in otl_counter.most_common(3)]) ) for pname, otl_counter in probe_offtarget_locations.items()
            )
        )

        #make a table
        df = pd.DataFrame(collections.OrderedDict((
            ('read_pairs_total', n_pairs),
            ('alignments_total', n_alignments),
            ('alignments_good', n_alignments_good),
            ('umis_good', n_umis_good),
            ('umis_good_coverage_ge_%d' % min_consensus_count, n_umis_good_coverage_ge_MIN),
            ('alignments_partial', n_alignments_partial),
            ('alignments_off_target', n_alignments_off_target),
            ('alignments_unmapped_single', n_alignments_unmapped_single),
            ('alignments_unmapped_both', n_alignments_unmapped_both),
            ('alignments_flagged', n_alignments_flagged),
            ('alignments_invalid_pair', n_alignments_invalid),
            ('alignments_mapq_lt_%d' % min_mapq, n_alignments_low_mapq),
            ('multimapping_alignments', n_alignments_multimappers),
            ('offtarget_locations', top_offtarget_locations)
            )))

        #fill with 0s and make ints
        df = pd.concat([df.loc[:, [c for c in df.columns if c != 'offtarget_locations']].fillna(0).astype(int), df['offtarget_locations']],
            axis=1)
        #set and sort by probe name (row index)
        df.index.rename('probe', inplace=True)
        df.sort_index(0, inplace=True)

        #ASSERTIONS
        #read_pairs_total should be == read_pairs_total from stats_overview
        assert df['read_pairs_total'].sum() == len(reads_processed) #each read pair should be in one probe row
        #assert df['alignments_total'].sum()*2 == n_rows #each alignment has two rows -- this falls apart if one mate is outside the target region!
        assert (df['read_pairs_total'] == df['alignments_total'] - df['multimapping_alignments']).all() #should have one non-multimapping alignment per pair
        assert (df['alignments_total'] == df[ [c for c in df.columns if c.startswith('alignments_') and c != 'alignments_total'] ].sum(axis=1)).all() #each alignment should be counted once in the alignment_XX columns

        df['fraction_pairs_good'] = 1.0 * df['alignments_good'] / df['read_pairs_total']

        if debug:
            print(df)
        else:
            outname = output_path
            df.to_csv(outname, index=True)
            print(outname)

def main():
    log.info('Called with arguments: "%s"', '" "'.join(sys.argv))

    #parse the arguments, which will be available as properties of args (e.g. args.probe)
    parser = argparse.ArgumentParser()
    #specify parameters
    parser.add_argument("--aggregate", help="folder with processed CSV files to aggregate")
    parser.add_argument("-d", "--design", help="CSV file with probes, arms sequences and locations")
    parser.add_argument("-i", "--input", help="input bam file name")
    parser.add_argument("-o", "--output", help="output file prefix + '.stats_alignment.csv'")
    parser.add_argument("--use-naive-groups", help="use UMIs from um tag instead of UG tags", action='store_true')
    parser.add_argument("--ignore-groups", help="ignore the UMI group entirely", action='store_true')
    parser.add_argument("--ignore-group-mismatch", help="ignore group mismatches (to deal with old alignments that have separate UMIs)", action='store_true')
    parser.add_argument("--min-consensus-count", help="minimum number of consistent read pairs supporting a UMI", default=2, type=int)
    parser.add_argument("--min-mapq", help="minimum mapping quality (remove read alignments with either read having mapq < X)", default=20, type=int)
    parser.add_argument("--include-primers", help="include primers when determining on/off target", action='store_true')

    parser.add_argument("--debug", help="enable debug output", action='store_true')
    args = parser.parse_args()

    if args.aggregate:
        aggregate(args.aggregate)
    else:
        process_file(args.design, args.input, args.output + '.stats_alignment.csv',
            args.min_mapq, args.min_consensus_count,
            args.include_primers, args.use_naive_groups, args.ignore_groups,
            args.ignore_group_mismatch,
            debug = args.debug)

    return 0

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    sys.exit(main())